"""
Ambient spaces of modular symbols. 
"""

#*****************************************************************************
#       SAGE: System for Algebra and Geometry Experimentation
#
#       Copyright (C) 2005 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

# Standard Python packages
import weakref

# SAGE packages
from   sage.misc.search import search

import sage.linalg.matrix_space as matrix_space
import sage.linalg.sparse_matrix as sparse
import sage.linalg.sparse_matrix_pyx as sparse_matrix_pyx
import sage.linalg.vector as vector
import sage.linalg.vector_space as vector_space
import sage.misc.misc as misc
import sage.modular.congroup as congroup
import sage.modular.dims as dims
import sage.modular.dirichlet as dirichlet
import sage.modular.hecke as hecke
import sage.rings.rational_field as rational_field
import sage.rings.rings as rings
import sage.rings.arith as arith
import sage.structure.function as function
import sage.structure.formal_sum as formal_sum

import boundary
import degenmap
import element
import heilbronn
import manin_symbols
import modular_symbols
import modsym
import p1list
import relation_matrix
import space
import subspace

_objsModularSymbolsAmbient = {}
class _uniqModularSymbolsAmbient(object):
    def __new__(cls, group=1, weight=2, sign=0,
                base_field=rational_field.RationalField(),
                character=None):
        key = (cls, group, weight, sign, base_field, character)
        if _objsModularSymbolsAmbient.has_key(key):
            x = _objsModularSymbolsAmbient[key]()
            if x != None:
                return x
        R = object.__new__(cls)
        _objsModularSymbolsAmbient[key] = weakref.ref(R)
        return R

class ModularSymbolsAmbient(_uniqModularSymbolsAmbient, 
                            space.ModularSymbolsSpace):
    """
    An ambient space of modular symbols for a congruence subgroup of SL_2(Z).
    
    This class is an abstract base class, so only derived classes should 
    be instantiated. 
    INPUT:
        weight -- int, the weight, must be >= 2
        group -- congroup.CongruenceGroup, a congruence subgroup. 
        sign -- int, either -1, 0, or 1
        base_field -- rings.Ring (defaults to the rational numbers)
    """
    def __init__(self, group=1, weight=2, sign=0,
                 base_field=rational_field.RationalField(),
                 character = None):
        """
        Initialize a space of modular symbols.
        """
        weight = int(weight)
        if weight <= 1:
            raise ValueError, "Weight (=%s) Modular symbols of weight <= 1 not defined."%weight
        if not isinstance(group, congroup.CongruenceSubgroup):
            raise TypeError, "group must be a congruence subgroup"

        sign = int(sign)
        if not isinstance(base_field, rings.Ring) and base_field.is_field():
            raise TypeError, "base_field must be of type Ring and be a field"

        if character == None and isinstance(group, congroup.Gamma0):
            character = dirichlet.TrivialCharacter(group.level(), base_field)

        space.ModularSymbolsSpace.__init__(self, group, weight, 
                       character, sign, base_field)

        formula = self._dimension_formula()
        if formula != None:
            d = self.dimension()
            assert d == formula, \
                   "Computed dimension (=%s) of ambient space \"%s\" doesn't match dimension formula (=%s)!\n"%(d, self, formula) + \
                   "ModularSymbolsAmbient: group = %s, weight = %s, sign = %s, base_field = %s, character = %s"%(
                         group, weight, sign, base_field, character)

    def manin_symbols(self):
        raise NotImplementedError

    def manin_generators(self):
        return self._manin_generators

    def manin_basis(self):
        try:
            return self._manin_basis
        except AttributeError:
            self.compute_presentation()
        return self._manin_basis

    def manin_relations(self):
        return self.__manin_relations

    def p1list(self):
        try:
            return self.__p1list
        except AttributeError:
            self.__p1list = p1list.P1List(self.level())
        return self.__p1list

    def relation_matrix(self):
        raise NotImplementedError

    def compute_presentation(self):
        B, basis, mod = relation_matrix.compute_presentation(
                self.manin_symbols(), self.sign(), 
                self.base_field(), self.weight())
        self._manin_generators = self.manin_symbols().manin_symbol_list()
        self._manin_basis = basis
        self._manin_gens_to_basis = B
        self._mod2term = mod
        
    def manin_gens_to_basis(self):
        try:
            return self._manin_gens_to_basis
        except AttributeError:
            self.compute_presentation()
            return self._manin_gens_to_basis            


    #####################################################################
    # Coercion
    #####################################################################
    def __call__(self, x, computed_with_hecke=False):
        r"""
        Coerce x into this modular symbols space (self). The result is
        either an element of self or a subspace of self.

        The allowed inputs for x are as follows:
        \begin{itemize}
        \item
            \class{Vector} -- a vector of the same degree, over the
                      same base field as self.  This defines the
                      corresponding linear combination of the basis of
                      self.

        \item

            \class{VectorSpace} -- returns the modular symbols
                           subspace of self defined defined by this
                           vector space.  The input vector space must
                           be over the same base field and be of the
                           same degree as the dimension of self.

        \item
            \class{list} -- list of elements of this modular symbols
                            space. returns the modular symbols
                            space generated by these elements. 

        \item
            \class{ManinSymbol} -- a Manin symbol of the same weight as the
                           space

        \item
            \class{ModularSymbolElement} -- a modular symbol whose
                                    ambient parent is this space of
                                    modular symbols. (TODO: make more
                                    sophisticated)
                                    
        \item
            0 -- the integer 0; results in the 0 modular symbol.

        \item
            3-tuple -- Given a 3-tuple (i,u,v), returns the modular
                       symbol in this space defined by the Manin
                       symbol $[X^{i}\cdot Y^{2-k-i}, (u,v)]$, where k is the
                       weight.  Note that we must have $0\leq i \leq 2-k$.
        \item
            2-tuple -- Given a 2-tuple (u,v), returns the modular
                       symbol defined by the Manin symbol
                       $[X^0 \cdot Y^{2-k}, (u,v)]$.
        \end{itemize}
        """
        if isinstance(x, vector.Vector):
            if x.degree() != self.dimension():
                raise TypeError, "Incompatible degrees: x has degree %s but modular symbols space has dimension %s"%(
                    x.degree(), self.dimension())
            if x.parent().base_field() != self.base_field():
                raise TypeError, "Vector x is over %s, but modular symbols space is over %s."%(
                    x.parent().base_field(), self.base_field())
            return element.ModularSymbolElement(self, x)
            
        elif isinstance(x, vector_space.VectorSpace):
            if x.degree() != self.dimension():
                raise TypeError, "Incompatible degrees: x has degree %s but modular symbols space has dimension %s"%(
                    x.degree(), self.dimension())
            if x.base_field() != self.base_field():
                raise TypeError, "Space x is over %s, but modular symbols space is over %s."%(
                    x.base_field(), self.base_field())
            space = subspace.ModularSymbolsSubspace(self, x)
            if computed_with_hecke:
                space._is_full_hecke_module = True
                space._is_anemic_hecke_module = True
            return space
            
        elif isinstance(x, (manin_symbols.ManinSymbol, element.ModularSymbolElement)):
            return self.element(x)

        elif isinstance(x, modular_symbols.ModularSymbol):
            return self(x.manin_symbol_rep())

        elif isinstance(x, (int, rings.Integer)) and x==0:
            return element.ModularSymbolElement(self, self.vector_space()(0))

        elif isinstance(x, tuple):
            if len(x) == 2:
                x = (0,x[0],x[1])
            if len(x) == 3:
                # Manin symbol of the form (i, u, v), which corresponds to [X^i*Y^(k-2-i), (u,v)].
                if x[0] < 0 or x[0] > self.weight()-2:
                    raise ValueError, "The first entry of the tuple (=%s) must be between 0 and k-2 (=%s)."%(
                        x, self.weight-2)
                y = manin_symbols.ManinSymbol(self.manin_symbols(), x)
                return self(y)

        elif isinstance(x, formal_sum.FormalSum):
            return misc.add([c*self(y) for c, y in x], self(0))

        elif isinstance(x, list):
            for v in x:
                if not (v in self):
                    raise TypeError, "each element of list must be in self"
            V = self.vector_space().subspace([v.vector() for v in x])
            return self(V)
            
        raise TypeError, "No coercion of %s into %s defined."%(x, self)

    def __getitem__(self, n):
        D = self.decomposition()
        if n < 0 or n >= len(D):
            raise IndexError, "index (=%s) must be between 0 and %s"%(n, len(D)-1)
        return D[n]

    def _action_on_modular_symbols(self, g):
        """
        Compute the matrix of the action of the 2x2 integer matrix g=[a,b,c,d]
        (which must be specified as an integer list) on self with respect
        to the standard basis.

        Use _matrix_of_operator_on_modular_symbols for more general operators.
        """
        if not isinstance(g, list):
            raise TypeError, "g must be a list"
        if not len(g) == 4:
            raise TypeError, "g must be a list of length 4"
        return self._matrix_of_operator_on_modular_symbols(self, [g])

    def __add__(self, other):
        if not isinstance(other, space.ModularSymbolsSpace):
            raise TypeError, "other (=%s) must be a ModularSymbolsSpace."%other
        if other.ambient_space() == self:
            return self
        raise ArithmeticError, "Sum only defined for subspaces of a common ambient modular symbols space."
        
    def _compute_hecke_matrix_prime_power(self, n, p, r):
        # convert input arguments to int's.
        (n,p,r) = (int(n), int(p), int(r))
        if not arith.is_prime(p):
            raise ArithmeticError, "p must be a prime"
        # T_{p^r} := T_p * T_{p^{r-1}} - eps(p)p^{k-1} T_{p^{r-2}}.
        pow = p**(r-1)
        if not self._hecke_matrices.has_key(pow):
            # The following will force computation of T_{p^s}
            # for all s<=r-1, except possibly s=0.
            self._compute_hecke_matrix(pow)
        if not self._hecke_matrices.has_key(1):
            self._compute_hecke_matrix(1)
        Tp = self._hecke_matrices[p]
        Tpr1 = self._hecke_matrices[pow]
        eps = self.character()
        k = self.weight()
        Tpr2 = self._hecke_matrices[pow/p]            
        self._hecke_matrices[n] = Tp*Tpr1 - eps(p)*(p**(k-1)) * Tpr2

    def _compute_hecke_matrix_general_product(self, n, F):
        n = int(n)
        prod = None
        for p, r in F:
            pow = int(p**r)
            if not self._hecke_matrices.has_key(pow):
                self._compute_hecke_matrix(pow)
            if prod == None:
                prod = self._hecke_matrices[pow]
            else:
                prod *= self._hecke_matrices[pow]
        self._hecke_matrices[n] = prod

    def _compute_dual_hecke_matrix(self, n):
        self._dual_hecke_matrices[n] = self.hecke_matrix(n).transpose()
    
    def _compute_hecke_matrix(self, n):
        n = int(n)
        if n<1:
            raise ValueError, "Hecke operator T_%s is not defined."%n
        if n==1:
            Mat = matrix_space.MatrixSpace(self.base_field(),self.dimension(), sparse=True)
            self._hecke_matrices[1] = Mat(1)
            return
        if arith.is_prime(n):
            self._compute_hecke_matrix_prime(n)
            return
        F = arith.factor(n)
        if len(F) == 1:  # nontrivial prime power case
            self._compute_hecke_matrix_prime_power(n, F[0][0],F[0][1])
            return
        else:
            self._compute_hecke_matrix_general_product(n, F)
            return

    def _compute_hecke_matrix_prime(self, p):
        """
        Compute and return the matrix of the p-th Hecke operator.
        
        EXAMPLES:
        We first compute some examples for Gamma0(N):
            
            sage: m = ModularSymbols(2, weight=4)
            sage: m._compute_hecke_matrix_prime(2).charpoly()
            x^2 - 9*x + 8
            
            sage: m = ModularSymbols(1,weight=12)
            sage: m._compute_hecke_matrix_prime(2).charpoly()
            x^3 - 2001*x^2 - 97776*x - 1180224
            sage: m._compute_hecke_matrix_prime(13).charpoly()
            x^3 - 1792159238562*x^2 - 2070797989680255444*x - 598189440899986203208472
            
            sage: m = ModularSymbols(1,weight=12, sign=-1)
            sage: m._compute_hecke_matrix_prime(5)
            [4830]
            sage: m._compute_hecke_matrix_prime(23)
            [18643272]
            
            sage: m = ModularSymbols(3,4)
            sage: m._compute_hecke_matrix_prime(2).charpoly()
            x^2 - 18*x + 81
            
            sage: m = ModularSymbols(6,4)
            sage: m._compute_hecke_matrix_prime(2).charpoly()
            x^6 - 14*x^5 + 29*x^4 + 172*x^3 - 124*x^2 - 320*x + 256
            sage: m._compute_hecke_matrix_prime(3).charpoly()
            x^6 - 50*x^5 + 511*x^4 + 3012*x^3 - 801*x^2 - 9234*x + 6561
            
            sage: m = ModularSymbols(15,4, sign=-1)
            sage: m._compute_hecke_matrix_prime(3).charpoly()
            x^4 - 2*x^3 + 18*x^2 + 18*x - 243
            
            sage: m = ModularSymbols(6,4)
            sage: m._compute_hecke_matrix_prime(7).charpoly()
            x^6 - 1344*x^5 + 666240*x^4 - 140462080*x^3 + 8974602240*x^2 + 406424518656*x + 3584872677376
            
            sage: m = ModularSymbols(4,4)
            sage: m._compute_hecke_matrix_prime(3).charpoly()
            x^3 - 84*x^2 + 2352*x - 21952
            
        We now compute some examples for modular symbols on Gamma1(N):
        
            sage: from sage.modular.congroup import Gamma1
            sage: m = ModularSymbols(Gamma1(13),2, sign=-1)
            sage: m._compute_hecke_matrix_prime(2).charpoly()
            x^2 + 3*x + 3
            
        The following is an example with odd weight:
        
            sage: m = ModularSymbols(Gamma1(5),3)
            sage: m._compute_hecke_matrix_prime(2).charpoly()
            x^4 - 10*x^3 + 50*x^2 - 170*x + 289
        
        This example has composite conductor and weight>2 dividing the
        conductor and nontrivial sign:
        
            sage: m = ModularSymbols(Gamma1(9),3, sign=1)
            sage: m._compute_hecke_matrix_prime(3).charpoly()
            x^6 + 3*x^4 - 19*x^3 + 24*x^2 - 9*x
        """
        p = int(p)
        try:
            return self._hecke_matrices[p]
        except AttributeError:
            self._hecke_matrices = {}
        except KeyError:
            pass
        tm = misc.verbose("Computing Hecke operator T_%s"%p)

        if arith.is_prime(p):
            H = heilbronn.HeilbronnCremona(p)
        else:
            H = heilbronn.HeilbronnMerel(p)

        B = self.manin_basis()
        cols = []
        N = self.level()
        mod2term = self._mod2term
        R = self.manin_gens_to_basis()
        K = self.base_field()
        W = R.new_matrix(nrows=len(B), ncols = R.nrows())
        syms = self.manin_symbols()
        n = len(syms)
        j = 0
        for i in B:
            for h in H:
                entries = syms.apply(i,h)
                for k, x in entries:
                    f, s = mod2term[k]
                    if s != 0:
                        W.set((j,f), W.get((j,f)) + s*K(x))
            j += 1
        tm = misc.verbose("start matrix multiply",tm)
        Tp = W*R
        misc.verbose("done matrix multiply",tm)
        self._hecke_matrices[p] = Tp
        misc.verbose("done making matrix",tm)
        return Tp


    def __heilbronn_operator(self, M, H, t=1):
        """
        Returns the matrix function from self to M defined by the pair (H, t),
        where H is a list of matrices and t is an integer.
        
        INPUT:
           self -- ModularSymbols , domain (an ambient space of modular symbols),
                                   
           M -- ModularSymbols, codomain (a space of modular symbols),

           H -- list, a list of matrices in M_2(Z),
           
           t  -- int, an integer.
           
        OUTPUT:
           MatrixFunction -- A function from self to M defined
                             by t and the matrices in H.
        """
        
        MS = matrix_space.MatrixSpace(self.base_field(), self.dimension(), M.dimension())
        if self.dimension() == 0 or M.dimension() == 0:
            A = MS(0)
            return function.MatrixFunction(A, self, M, "Heilbronn operator(%s,%s)"%(H,t))
        
        rows = []
        B = self.manin_basis()
        syms = self.manin_symbols()
        k = self.weight()
        for n in B:
            z = M(0)
            i, u, v = syms[n]
            # We apply each Heilbronn matrix to the
            #    Manin symbol [X^i*Y^(k-2-i), (u,v)]
            for h in H:
                # Apply h to the polynomial part
                (a,b,c,d) = tuple(h)
                # P gives the ordered coefficients of (a*X+b*Y)^i*(c*X+d*Y)^(j-i)
                P = manin_symbols.apply_to_monomial(i, k-2, a,b,c,d)
                # Apply h to the (u,v) part of the Manin symbol
                (uu,vv) = (u*a+v*c, u*b+v*d)

                # For the generalized Heilbronn operator, we through away any
                # symbols for which the (u,v) part of the symbol doesn't have
                # both entries divisible by t.
                if t != 1:
                    if uu%t != 0 or vv%t != 0:
                        continue
                    uu = uu//t
                    vv = vv//t

                # Now coerce each Manin symbol
                #
                #         P[m]*[X^m*Y^(k-2-m), (uu,vv)],    for m=0,...,len(P)
                #
                # into the image space M and add that to z.
                # Note that we coerce in Manin symbols as tuples.
                for m in range(len(P)):
                    x = M((m,uu,vv))
                    z += x*P[m]
                
            rows.append(z.vector())
            
        A = MS(rows)
        return function.MatrixFunction(\
                A, self, M, "Heilbronn operator(%s,%s)"%(H,t))

    def __len__(self):
        return len(self.decomposition())

    def __repr__(self):
        return "Space of Modular Symbols of dimension %s and weight %s for %s with sign %s and character %s over %s"%(
                self.dimension(), self.weight(), self.group(), self.sign(), self.character(), self.base_field())

    def _hecke_image_of_ith_basis_vector(self, n, i):
        """
        Return the image under the Hecke operator T_n of the ith basis vector.
        """
        return self.hecke_operator(n)(self.gen(i))

    def _matrix_of_operator_on_modular_symbols(self, codomain, R):
        """
        INPUT:
            self -- this space of modular symbols
            codomain -- space of modular symbols
            R -- list of lists [a,b,c,d] of length 4, which we view as elements of GL_2(Q).
            
        OUTPUT:
            a matrix, which represents the operator
            $$
               x \mapsto \sum_{g in R} g.x
            $$
            where g.x is the formal linear fractional transformation on modular symbols.
        """
        rows = []
        for b in self.basis():
            v = formal_sum.FormalSum(0, check=False)
            for c, x in b.modular_symbol_rep():
                for g in R:
                    y = x.apply(g)
                    v += y*c
            w = codomain(v).vector()
            rows.append(w)
        M = matrix_space.MatrixSpace(self.base_field(), len(rows), codomain.degree(), sparse=False)
        return M(rows)

    def _set_dual_vector_space(self, V):
        pass  # setting dual vector space of ambient space is ignored

    def ambient_space(self):
        return self

    def ambient_vector(self, v):
        """
        Given a vector v in self.vector_space(), return the
        corresponding element in the vector space of the ambient
        modular symbols space.
        """
        return v

    def atkin_lehner_operator(self, d):
        """
        Return the Atkin-Lehner operator W_d on this space, where d is
        a divisor of the level N such that N/d and d are coprime.
        """
        d = int(d)
        if self.level() % d != 0:
            raise ArithmeticError, "d (=%s) must be a divisor of the level (=%s)"%(d,self.level())

        N = self.level()
        for p, e in arith.factor(d):
            v = arith.valuation(N, p)
            if e < v:
                d *= p**(v-e)
        d = int(d)
        try:
            return self.__atkin_lehner_operator[d]
        except AttributeError:
            self.__atkin_lehner_operator = {}
        except KeyError:
            pass
        
        k = self.weight()
        R = self.base_ring()
        
        g, x, y = arith.xgcd(d, -N//d)
        g = [d*x, y, N, d]
        A = self._action_on_modular_symbols(g)
        scale = R(d)**(1 - k//2)
        Wmat = scale * A
        W = function.MatrixFunction(Wmat, self, self, "Atkin-Lehner operator W_%s"%d)
        self.__atkin_lehner_operator[d] = W
        return W

    def boundary_map(self):
        """
        The boundary map to the corresponding space of boundary
        modular symbols.
        """
        try:
            return self.__boundary_map
        except AttributeError:
            # compute boundary map 
            B = self.boundary_space()
            I = [B(b) for b in self.basis()]
            V = B.vector_space()
            W = matrix_space.MatrixSpace(self.base_field(), len(I), V.degree(), sparse=True)
            A = W([x.vector() for x in I])
            self.__boundary_map = function.MatrixFunction(A, self, B, "Boundary Map")
            return self.__boundary_map

    def boundary_space(self):
        raise NotImplementedError
        
    def commit(self):
        """
        Try to commit this modular symbols space to the database.
        """
        if self.__use_db:
            db.save(self)

    def complement(self):
        """
        Return the largest Hecke-stable complement of this space.
        """
        return self.zero_subspace()

    def cuspidal_subspace(self):
        """
        The cuspidal subspace.
        """
        try:
            return self.__cuspidal_subspace
        except AttributeError:
            S = self.boundary_map().kernel()
            S._is_full_hecke_module = True
            if self.base_ring().characteristic() == 0:
                d = self._cuspidal_subspace_dimension_formula()
                assert d == S.dimension(), "According to dimension formulas the cuspidal subspace of \"%s\" has dimension %s; however, computing it using modular symbols we obtained %s, so there is a bug (please report!)."%(self, d, S.dimension())
            self.__cuspidal_subspace = S            
        return self.__cuspidal_subspace

    def _degeneracy_matrix_raising(self, level):
        raise NotImplementedError

    def degeneracy_map(self, level, t=1):
        """
        The t-th degeneracy map from self to modular symbols of the
        given level.  The level of self must be a divisor or multiple
        of level, and t must be a divisor of the quotient.             
        
        NOTE: We use Proposition 2.6.15 in Merel's 1585 paper (or
              Prop 15 in the dvi version of that paper). 
        INPUT:
            level -- int, the level of the codomain of the map (positive int).
            t  -- int, the parameter of the degeneracy map, i.e., the map is
                  related to f(q) |--> f(q^t).
                  
        OUTPUT:
            A linear function from self to the space of modular symbols
            of given level with the same weight, character, sign,
            etc., as this space.
            
        EXAMPLES:
            
        """
        level = int(level); t = int(t)
        
        err = False
        if self.level() % level == 0:
            quo = self.level() // level
            if quo % t != 0:
                err = True
        elif level % self.level() == 0:
            quo = level // self.level() 
            if quo % t != 0:
                err = True
        else:
            err = True
        if err:
            raise ValueError, ("The level of self (=%s) must be a divisor or multiple of " + \
                               "level (=%s), and t (=%s) must be a divisor of the quotient (=%s).")%\
                               (self.level(), level, t, quo)
        
        key = (level,t)
        try:
            self._degeneracy_maps
        except AttributeError:
            self._degeneracy_maps = {}
            
        if self._degeneracy_maps.has_key(key):
            return self._degeneracy_maps[key]

        eps = self.character()
        if not (eps is None) and level % eps.conductor() != 0:
            raise ArithmeticError, "The conductor of the character of this space " + \
                  "(=%s) must be divisible by level (=%s)."%\
                  (eps.conductor(), level)

        M = self.modular_symbols_of_level(level)

        if M.dimension() == 0:
            
            A = matrix_space.MatrixSpace(self.base_field(), self.dimension(),0)(0)

        elif self.level() % level == 0:  # lower the level

            # Use Proposition 2.6.15 in Merel's 1585 paper (or Prop 15 in
            # electronic version of that paper).
            H = heilbronn.HeilbronnMerel(t)
            A = self.__heilbronn_operator(M,H,t).matrix()
            
        elif level % self.level() == 0:  # raise the level
            # Use modular symbols.   Merel also has a Manin symbols
            # formula in his paper 
            # TODO: Implement the other Manin symbols formula and
            # compare timings.  It should be faster.
            if t != 1:   
                d1 = self.degeneracy_map(level, 1).matrix()
                T = M.hecke_matrix(t)
                A = (1/self.base_ring()(t)) * d1 * T
            else:
                A = self._degeneracy_matrix_raising(M)

        d = degenmap.DegeneracyMap(A, self, M, t)
        self._degeneracy_maps[key] = d
        return d

    def dimension(self):
        """
        Returns the dimension of self.
        
        INPUT:
           ModularSymbols self -- arbitrary space of modular symbols
           
        OUTPUT:
           int -- the dimension
           
        ALGORITHM:
           If self is a subspace, then it is defined by a vector
           space, and the dimension is known.  If self is ambient,
           this function computes a presentation for self, and return
           the number of free generators in the presentation.  The
           dimension could also be computed in many cases using
           dimension formulas, but we don't do this.
           
        EXAMPLES:
            sage: M = ModularSymbols(389)
            sage: M.dimension()
            65
        """
        try:
            return self.__dimension
        except AttributeError:
            self.__dimension = len(self.manin_basis())
        return self.__dimension

    def eisenstein_subspace(self):
        """
        Return the Eisenstein subspace of this space of modular symbols.
        """
        try:
            return self.__eisenstein_subspace
        except AttributeError:
            self.__eisenstein_subspace = self.cuspidal_subspace().complement()
            return self.__eisenstein_subspace
        
            #if not self.is_ambient():
            #    E = self.ambient_space().eisenstein_subspace()
            #    self.__eisenstein_subspace = self.intersection(E)
            #    return self.__eisenstein_subspace
            #D = self.decomposition()
            #E = [A for A in D if not A.is_cuspidal()]
            #V = misc.add(E, self.zero_subspace())
            #self.__eisenstein_subspace = V
            #return V
            

    def embedded_vector_space(self):
        """
        Returns the vector subspace of the vector space of the ambient
        space that defines self as a subspace of modular symbols.
        
        OUTPUT:
            VectorSpace -- a vector space.
        """
        return self.vector_space()

    def embedded_dual_vector_space(self):
        return self.vector_space()

    def element(self, x):
        """
        Creates and returns an element of self from a modular symbol,
        if possible.
        
        INPUT:
           x -- an object of one of the following types:
                ModularSymbol, ManinSymbol.

        OUTPUT:
        
           ModularSymbol -- a modular symbol with parent self.
           
        """
        if isinstance(x, manin_symbols.ManinSymbol):
            if not x.parent().weight() == self.weight():
                raise ArithmeticError, "incompatible weights: Manin symbol has weight %s, but modular symbols space has weight %s"%(
                    x.parent().weight(), self.weight())
            t = self.manin_symbols().index(x.tuple())
            if isinstance(t, tuple):
                i, scalar = t
                v = self.manin_gens_to_basis()[i] * scalar
            else:
                v = self.manin_gens_to_basis()[t]
            return element.ModularSymbolElement(parent=self, vector=v)
        
        elif isinstance(x, element.ModularSymbolElement):
            M = x.parent()
            if M.ambient_space() != self:
                # TODO -- sometimes do something more sophisticated here.
                raise TypeError, "Modular symbol (%s) does not lie in this space."%x
            return self(x.vector())
            
        else:
            raise ValueError, "Cannot create element of %s from %s."%(x,self)

    def filename(self):
        """
        Returns the filename of self that should be used to store self
        in the database.
        INPUT:
           self -- space of modular symbols
        OUTPUT:
           str -- a string
        """
        return "modsym-%s_%s_%s_%s_%s"%\
               (self.__group.name(),self.__weight,self.__character,\
                self.__sign,self.__base_field.name())

    def decomposition_matrix(self):
        """
        Returns the matrix whose columns form a basis for the
        canonical sorted decomposition of self coming from the Hecke
        operators.
        
        If the simple factors are D_0,...,D_n, then the first few
        columns are an echelon basis for D_0, the next an echelon
        basis for D_1, the next a basis for D_2, etc.
        """
        try:
            return self.__decomposition_matrix_cache
        except AttributeError:
            rows = []
            for A in self.decomposition():
                for x in A.basis():
                    rows.append(x.vector())
            A = matrix_space.MatrixSpace(self.base_field(),self.dimension())(rows)
            self.__decomposition_matrix_cache = A
            return self.__decomposition_matrix_cache

    def decomposition_matrix_inverse(self):
        """
        Returns the inverse of the matrix returned by decomposition_matrix().
        """
        try:
            return self.__decomposition_matrix_inverse_cache
        except AttributeError:
            self.__decomposition_matrix_inverse_cache = ~self.decomposition_matrix()
            return self.__decomposition_matrix_inverse_cache
    
    def dual_star_involution_matrix(self):
        """
        Return the matrix of the dual star involution, which is
        induced by complex conjugation on the linear dual of modular
        symbols.
        """
        try:
            return self.__dual_star_involution_matrix
        except AttributeError:
            pass
        self.__dual_star_involution_matrix = self.star_involution().matrix().transpose()
        return self.__dual_star_involution_matrix

    def hecke_algebra(self):
        """
        Returns the Hecke algebra associated to self.
        
        INPUT:
           ModularSymbols -- Hecke stable space of modular symbols
           
        OUTPUT:
           hecke.HeckeAlgebra -- the Hecke algebra attached to self.
           
        ALGORITHM:
           Creates the hecke.HeckeAlgebra attached to self.  For how
           that works, see the documentation of the hecke module.
        """
        try:
            return self.__hecke_algebra
        except AttributeError:
            self.__hecke_algebra = hecke.HeckeAlgebra(self)
        return self.__hecke_algebra

    def hecke_bound(self):
        print "WARNING: ambient.py -- hecke_bound; returning unproven guess."
        return 3*self.sturm_bound() + 10
        
    def intersection(self, other):
        """
        Returns the intersection of self and other, which must both
        lie in a common ambient space of modular symbols.
        
        EXAMPLES:
            sage: M = ModularSymbols(43, sign=1)
            sage: A = M[0] + M[1]
            sage: B = M[1] + M[2]
            sage: A.dimension(), B.dimension()
            (2, 3)
            sage: C = A.intersection(B); C.dimension()  # TODO
            1
        """
        if not isinstance(other, space.ModularSymbolsSpace):
            raise TypeError, "other (=%s) must be a ModularSymbolsSpace."%other
        if self.ambient_space() != other.ambient_space():
            raise ArithmeticError, "Intersection only defined for subspaces of a common ambient modular symbols space."
        return other  # since self is ambient, so the intersection must equal other.

    def is_ambient(self):
        """
        Returns true if and only if self is an ambient space of
        modular symbols.
        command.
        
        WARNING: self can be ambient by virtue of being equal to an
        ambient space, even if it is not of type
        ModularSymbolsAmbient.  For example, decomposing a simple
        ambient space yields a single factor, and that factor *is*
        also considered an ambient space.
        
        EXAMPLES:
            sage: m = ModularSymbols(10)
            sage: m.is_ambient()
            True
            
            sage: a = m[0]  # the unique simple factor
            sage: a == m
            True
            sage: a.is_ambient()
            True
        """
        return True

    def is_anemic_hecke_module(self):
        return True

    def is_full_hecke_module(self):
        return True
    
    def is_cuspidal(self):
        try:
            return self.__is_cuspidal
        except AttributeError:
            S = self.ambient_space().cuspidal_subspace()
            self.__is_cuspidal = (S.dimension() == self.dimension())
        return self.__is_cuspidal

    def is_eisenstein(self):
        try:
            return self.__is_eisenstein
        except AttributeError:
            S = self.ambient_space().eisenstein_subspace()
            self.__is_eisenstein = self.is_subspace(S)
        return self.__is_eisenstein

    def is_new(self, p=None):
        try:
            if self.__is_new.has_key(p):
                return self.__is_new[p]
        except AttributeError:
            pass
        self.new_subspace(p)
        return self.__is_new[p]

    def is_subspace(self, V):
        """
        Returns True if and only if self is a subspace of V.
        """
        if not V.is_ambient():
            return False
        return V.ambient_space() == self

    def linear_combination_of_basis(self, v):
        return self(v)

    def load(self):
        """
        Try to load the ambient space self from the database.
        INPUT:
           ModularSymbols self -- an ambient space of modular symbols
        OUTPUT:
           bool --  True on success and False on failure.
        """
        try:
            x = db.load(self)
        except:     # todo -- make more precise
            return False
        if x != None:
            # todo -- need to loop through the attributes of the
            # object and load them here!!!  make a list of exceptions...
            # In particular, C++ level attributes have to be
            # handled carefully.
            self.__presentation = x.__presentation
            self.__dimension = x.__dimension
            self.__basis = x.__basis
            self._ModularSymbolsSpace__hecke_matrices = x._ModularSymbolsSpace__hecke_matrices
            x.__use_db = False # so doesn't save self, which would waste time
            return True
        else:
            return False

    def manin_symbols_basis(self):
        """
        A list of Manin symbols that form a basis for the ambient
        space self.
        INPUT:
           ModularSymbols self -- an ambient space of modular symbols
        OUTPUT:
           list -- a list of 2-tuples (if the weight is 2) or 3-tuples, 
                   which represent the Manin symbols basis for self.
        EXAMPLES:
            sage: m = ModularSymbols(23)
            sage: m.manin_symbols_basis()
            [(1,0), (1,17), (1,19), (1,20), (1,21)]
            sage: m = ModularSymbols(6, weight=4, sign=-1)
            sage: m.manin_symbols_basis()
            [[X^2,(2,1)]]
        """
        s = self.manin_symbols()
        return [s.manin_symbol(i) for i in self.manin_basis()]

    def modular_symbols_of_level(self, N):
        """
        Returns a space of modular symbols with the same defining
        properties (weight, sign, etc.) as this space except with the
        level N.

        For example, if self is the space of modular symbols of weight
        2 for Gamma_0(22), and level is 11, then this function returns
        modular symbols of weight 2 for Gamma_0(11).
        """
        raise NotImplementedError

    def _compute_sign_subspace(self, sign, compute_dual=True):
        """
        Return the subspace of self that is fixed under the star involution.

        INPUT:
            sign -- int (either -1 or +1)
            compute_dual -- bool (default: True) also compute dual subspace.
                            This are useful for many algorithms.
        OUTPUT:
            subspace of modular symbols

        """
        S = self.star_involution().matrix() - sign
        P = self(S.kernel())
        if compute_dual:
            P._set_dual_vector_space(S.transpose().kernel())
        return P
    
    def new_subspace(self, p=None):
        """
        Returns the new or p-new subspace of self.
        
        INPUT:
            p -- (default: None); if not None, return only the p-new subspace.
            
        OUTPUT:
            ModularSymbols -- the new or p-new subspace of self
            
        EXAMPLES:
            sage: m = ModularSymbols(33); m.dimension()
            9
            sage: m.new_subspace().dimension()
            3
            sage: m.new_subspace(3).dimension()
            4
            sage: m.new_subspace(11).dimension()
            8
        """
        try:
            if self.__is_new[p]:
                return self
        except AttributeError:
            self.__is_new = {}
        except KeyError:
            pass

        if self.dimension() == 0:
            self.__is_new[p] = True
            return self
        try:
            return self.__new_subspace[p]
        except AttributeError:
            self.__new_subspace = {}
        except KeyError:
            pass

        # Construct the degeneracy map d.
        N = self.level()
        d = None
        eps = self.character()
        if eps == None:
            f = 1
        else:
            f = eps.conductor()
        if p == None:
            D = arith.prime_divisors(N)
        else:
            if N % p != 0:
                raise ValueError, "p must divide the level."
            D = [p]
        for q in D:
            if ((N//q) % f) == 0:
                NN = N//q
                d1 = self.degeneracy_map(NN,1).matrix()
                if d is None:
                    d = d1
                else:
                    d = d.augment(d1)
                d = d.augment(self.degeneracy_map(NN,q).matrix())
            #end if
        #end for
        if d is None or d == 0:
            self.__is_new[p] = True
            return self
        else:
            self.__is_new[p] = False
        ns = self(d.kernel())
        ns.__is_new = {p:True}
        ns._is_full_hecke_module = True
        self.__new_subspace[p] = ns
        return self.__new_subspace[p]

    def star_involution(self):
        """
        Return the star involution on self, which is induced by complex
        conjugation on modular symbols.
        """
        try:
            return self.__star_involution
        except AttributeError:
            pass
        S = self.__heilbronn_operator(self, [[-1,0, 0,1]], 1)
        S.name("Star involution on %s"%self)
        self.__star_involution = S
        return self.__star_involution

        
    def subspace_generated_by_images(self, M):
        """
        Return the subspace of this ambient modular symbols space
        generated by the images under all degeneracy maps of M.  The
        space M must have the same weight, sign, and group or
        character as this space.
        """
        S = M.zero_subspace()
        if self.level() % M.level() == 0:
            for t in arith.divisors(self.level() // M.level()):
                d = M.degeneracy_map(self.level(), t)
                if d.codomain() != self:
                    raise ArithmeticError, "incompatible spaces of modular symbols"
                S += d.image()
                
        if self.is_full_hecke_module(compute=False):
            S._is_full_hecke_module = True
        elif self.is_anemic_hecke_module():
            S._is_anemic_hecke_module = True
            
        return S

    def fcp(self, n):
        """
        Returns the factorization of the characteristic polynomial of
        the Hecke operator T_n of index n.
        
        INPUT:
           ModularSymbols self -- space of modular symbols invariant
                                   under the Hecke operator of index n.
           int n -- a positive integer.
           
        OUTPUT:
           list -- list of the pairs (g,e), where g is an irreducible
                   factor of the characteristic polynomial of T_n, and
                   e is its multiplicity.
                   
        ALGORITHM:
           Call hecke_operator to compute the hecke operator, call
           charpoly on that, and call factor on the result of
           charpoly.  Thus the implementation is just a one-line call
           to other functions.
           
        EXAMPLES:
            sage: m = ModularSymbols(23, 2, sign=1)
            sage: m.fcp(2)
            [(x - 3, 1), (x^2 + x - 1, 1)]
            sage: m.hecke_operator(2).charpoly().factor()
            [(x - 3, 1), (x^2 + x - 1, 1)]
        """
        n = int(n)
        if n <= 0:
            raise ArithmeticError, "n (=%s) must be positive"%n
        return self.hecke_operator(n).charpoly().factor()

    def use_db(self):
        """
        Returns true if self is loaded and saved to disk.
        INPUT:
           ModularSymbols self -- an ambient space of modular symbols
        OUTPUT
           bool
        """
        return self.__use_db



class ModularSymbolsAmbient_wtk_g0_Q(ModularSymbolsAmbient):
    """
    Modular symbols for Gamma_0(N) of integer weight k > 2 over the field Q.
    """
    def __init__(self, N, k, sign=0):
        """
        Initialize a space of modular symbols of weight k for Gamma_0(N), over Q.
        
        For weight 2, it is faster to use ModularSymbols_wt2_g0_Q.

        INPUT:
            N -- int, the level
            k -- integer weight >= 2.
            sign -- int, either -1, 0, or 1
        EXAMPLES:
            sage: ModularSymbols(1,12)
            Full Modular Symbols space for Gamma_0(1) of weight 12 with sign 0 and dimension 3 over Rational Field
            sage: ModularSymbols(1,12, sign=1).dimension()
            2
            sage: ModularSymbols(15,4, sign=-1).dimension()
            4
            sage: ModularSymbols(6,6).dimension()
            10
            sage: ModularSymbols(36,4).dimension()
            36
        """
        N = int(N)
        k = int(k)
        sign = int(sign)
        if not sign in [-1,0,1]:
            raise TypeError, "sign must be an int in [-1,0,1]"
            
        ModularSymbolsAmbient.__init__(self, weight=k, group=congroup.Gamma0(N),
                            sign=sign, base_field=rational_field.RationalField())


    def _dimension_formula(self):
        N, k, sign = self.level(), self.weight(), self.sign()
        if sign != 0: return None
        if k%2 == 1:
            return 0
        elif k > 2:
            return 2*dims.dimension_cusp_forms_gamma0(N,k) + dims.c0(N)
        else:
            return 2*dims.dimension_cusp_forms_gamma0(N,k) + dims.c0(N)-1

    def __repr__(self):
        return ("Full Modular Symbols space for Gamma_0(%s) of weight %s with sign %s " + \
                "and dimension %s over %s")%(self.level(),self.weight(), self.sign(), 
                    self.dimension(), self.base_field())

    def _cuspidal_subspace_dimension_formula(self):
        N, k, sign = self.level(), self.weight(), self.sign()        
        if sign == 0:
            m = 2
        else:
            m = 1
        return m * dims.dimension_cusp_forms_gamma0(N, k)

    def _degeneracy_matrix_raising(self, M):
        R = congroup.degeneracy_coset_representatives_gamma0(M.level(), self.level(), 1)
        return self._matrix_of_operator_on_modular_symbols(M, R)

    def _cuspidal_new_subspace_dimension_formula(self):
        N, k, sign = self.level(), self.weight(), self.sign()        
        if sign == 0:
            m = 2
        else:
            m = 1
        return m * dims.dimension_new_cusp_forms_gamma0(N, k)

    def boundary_space(self):
        try:
            return self.__boundary_space
        except AttributeError:
            pass
        self.__boundary_space = boundary.BoundarySymbols_wtk_g0_Q(
            self.level(), self.weight(), self.sign())
        return self.__boundary_space

    def manin_symbols(self):
        try:
            return self.__manin_symbols
        except AttributeError:
            self.__manin_symbols = manin_symbols.ManinSymbolList_gamma0(
                level=self.level(), weight=self.weight())
        return self.__manin_symbols

    def modular_symbols_of_level(self, N):
        """
        Returns a space of modular symbols with the same defining
        properties (weight, sign, etc.) as this space except with the
        level N.

        For example, if self is the space of modular symbols of weight
        2 for Gamma_0(22), and level is 11, then this function returns
        modular symbols of weight 2 for Gamma_0(11).
        """
        if self.weight() == 2:
            return ModularSymbolsAmbient_wt2_g0_Q(N, self.sign())
        else:
            return ModularSymbolsAmbient_wtk_g0_Q(N, self.weight(),self.sign())

    def modular_symbols_of_sign(self, sign):
        """
        Returns a space of modular symbols with the same defining
        properties (weight, sign, etc.) as this space except with
        given sign.
        """
        sign = int(sign)
        if not (sign in [0,-1,1]):
            raise ValueError, "sign (=%s) must be an int in [0,-1,1]."%sign
        if self.weight() == 2:
            return ModularSymbolsAmbient_wt2_g0_Q(self.level(), sign)
        else:
            return ModularSymbolsAmbient_wtk_g0_Q(self.level(), self.weight(), sign)

    def modular_symbols_of_weight(self, k):
        """
        Returns a space of modular symbols with the same defining
        properties (weight, sign, etc.) as this space except with
        weight k.
        """
        if k == 2:
            return ModularSymbolsAmbient_wt2_g0_Q(self.level(), self.sign())
        else:
            return ModularSymbolsAmbient_wtk_g0_Q(self.level(), k, self.sign())

    
class ModularSymbolsAmbient_wt2_g0_Q(ModularSymbolsAmbient_wtk_g0_Q):
    """
    Modular symbols for Gamma_0(N) of integer weight 2 over the field Q.
    """
    def __init__(self, N, sign=0):
        """
        Initialize a space of modular symbols.
        INPUT:
            N -- int, the level
            sign -- int, either -1, 0, or 1
        OUTPUT:
            The space of modular symbols of weight 2, trivial character,
            level N and given sign.
        """            
        ModularSymbolsAmbient_wtk_g0_Q.__init__(self, 
                N=N, k=2, sign=sign)

    def _dimension_formula(self):
        N, sign = self.level(), self.sign()                
        if sign != 0: return None
        return 2*dims.dimension_cusp_forms_gamma0(N,2) + dims.c0(N) - 1

    def _cuspidal_subspace_dimension_formula(self):
        if self.sign() == 0:
            m = 2
        else:
            m = 1
        return m * dims.dimension_cusp_forms_gamma0(self.level(), 2)

    def _cuspidal_new_subspace_dimension_formula(self):
        if self.sign() == 0:
            m = 2
        else:
            m = 1
        return m * dims.dimension_new_cusp_forms_gamma0(self.level(), 2)

                    
    def _compute_hecke_matrix_prime(self, p):
        """
        Compute and return the matrix of the p-th Hecke operator.
        EXAMPLES:
            sage: m = ModularSymbols(37,2)
            sage: m._compute_hecke_matrix_prime(2).charpoly()
            x^5 + x^4 - 8*x^3 - 12*x^2
        """
        assert arith.is_prime(p), "p must be prime."
        try:
            return self._hecke_matrices[p]
        except AttributeError:
            self._hecke_matrices = {}
        except KeyError:
            pass
        tm = misc.verbose("Computing Hecke operator T_%s"%p)
        
        H = heilbronn.HeilbronnCremona(p)
        ##H = heilbronn.HeilbronnMerel(p)
        B = self.manin_basis()
        cols = []
        N = self.level()
        P1 = self.p1list()
        mod2term = self._mod2term
        R = self.manin_gens_to_basis()
        W = R.new_matrix(nrows=len(B), ncols = R.nrows())  # the 0 with given number of rows and cols.
        j = 0
        tm = misc.verbose("Matrix non-reduced matrix W", tm)
        for i in B:
            # The following step is where most of the time is spent.
            c,d = P1[i]
            v = H.apply(c,d, N)
            # v is now a list of pairs ((c,d),m), where m is the
            # number of times that (c,d) appears in the image of x
            # under the matrices in H.  Also, the pairs (c,d) are
            # normalized.
            # Let ind(c,d) denote the index of the normalized pair
            # (c,d) in the fixed ordered list of elements of
            # P1(Z/NZ).  Then the list of pairs (ind(c,d), m)
            # obtained from the above list defines a sparse vector
            # s, and the image of x under T_p is the product
            # of s with the matrix R defined above.
            for z, m in v:
                k = P1.index_of_normalized_pair(z[0],z[1])
                if k != -1:
                    f, s = mod2term[k]
                    if s != 0:
                        W.set((j,f), W.get((j,f)) + s*m)
            j += 1
        tm = misc.verbose("done making non-reduced matrix",tm)
        misc.verbose("start sparse matrix-matrix multiply")
        Tp = W*R
        misc.verbose("done multiplying",tm)
        self._hecke_matrices[p] = Tp
        misc.verbose("done making matrix",tm)
        return Tp

    def boundary_space(self):
        try:
            return self.__boundary_space
        except AttributeError:
            pass
        self.__boundary_space = boundary.BoundarySymbols_wtk_g0_Q(
            self.level(), self.weight(), self.sign())
        return self.__boundary_space
    
        
class ModularSymbolsAmbient_wtk_g1_Q(ModularSymbolsAmbient):
    def __init__(self, level, weight, sign=0):
        """
        Initialize a space of modular symbols for Gamma1(N).
        INPUT:
            level -- int, the level
            weight -- int, the weight >= 2
            sign -- int, either -1, 0, or 1
        EXAMPLES:
            sage: ModularSymbolsAmbient_wtk_g1_Q(17,2)
            Full Modular Symbols space for Gamma_1(17) of weight 2 with sign 0 and dimension 25 over Rational Field
            sage: [ModularSymbolsAmbient_wtk_g1_Q(7,k).dimension() for k in [2,3,4,5]]
            [5, 8, 12, 16]
        """            
        ModularSymbolsAmbient.__init__(self, 
                weight=weight, 
                group=congroup.Gamma1(level),
                sign=sign, 
                base_field=rational_field.RationalField())

    def _dimension_formula(self):
        level, weight, sign = self.level(), self.weight(), self.sign()
        if sign != 0: return None
        d = 2*dims.dimension_cusp_forms_gamma1(level,weight) + dims.c1(level)
        if level == 1 and weight%2 == 1:
            return 0
        if weight == 2:
            return d - 1
        if weight % 2 == 0:
            return d

        # TODO: I don't know a formula for dim M_k(Gamma_1(N)) for odd k!!!
        
        return None  
                
    def __repr__(self):
        return ("Full Modular Symbols space for Gamma_1(%s) of weight %s with sign %s " + \
                "and dimension %s over %s")%(self.level(),self.weight(), self.sign(),
                    self.dimension(), self.base_field())
 
    def _cuspidal_subspace_dimension_formula(self):
        if self.sign() == 0:
            m = 2
        else:
            m = 1
        return m * dims.dimension_cusp_forms_gamma1(self.level(), self.weight())

    def _cuspidal_new_subspace_dimension_formula(self):
        if self.sign() == 0:
            m = 2
        else:
            m = 1
        return m * dims.dimension_new_cusp_forms_gamma1(self.level(), self.weight())
        
                    
    def _compute_hecke_matrix_prime_power(self, n, p, r):
        return self._compute_hecke_matrix_prime(n)
 
    def _degeneracy_matrix_raising(self, M):
        R = congroup.degeneracy_coset_representatives_gamma1(M.level(), self.level(), 1)
        return self._matrix_of_operator_on_modular_symbols(M, R)

    def boundary_space(self):
        try:
            return self.__boundary_space
        except AttributeError:
            pass
        self.__boundary_space = boundary.BoundarySymbols_wtk_g1_Q(
            self.level(), self.weight(), self.sign())
        return self.__boundary_space

    def manin_symbols(self):
        try:
            return self.__manin_symbols
        except AttributeError:
            self.__manin_symbols = manin_symbols.ManinSymbolList_gamma1(
                level=self.level(), weight=self.weight())
        return self.__manin_symbols

        
    def modular_symbols_of_level(self, N):
        """
        Returns a space of modular symbols with the same defining
        properties (weight, sign, etc.) as this space except with the
        level N.

        For example, if self is the space of modular symbols of weight
        2 for Gamma_0(22), and level is 11, then this function returns
        modular symbols of weight 2 for Gamma_0(11).
        """
        return ModularSymbolsAmbient_wtk_g1_Q(N, self.weight(),self.sign())

    def modular_symbols_of_sign(self, sign):
        """
        Returns a space of modular symbols with the same defining
        properties (weight, sign, etc.) as this space except with
        given sign.
        """
        sign = int(sign)
        if not (sign in [0,-1,1]):
            raise ValueError, "sign (=%s) must be an int in [0,-1,1]."%sign
        return ModularSymbolsAmbient_wtk_g1_Q(self.level(), self.weight(), sign)

    def modular_symbols_of_weight(self, k):
        """
        Returns a space of modular symbols with the same defining
        properties (weight, sign, etc.) as this space except with
        weight k.
        """
        return ModularSymbolsAmbient_wtk_g1_Q(self.level(), k, self.sign())



class ModularSymbolsAmbient_wtk_eps(ModularSymbolsAmbient):
    def __init__(self, eps, weight, sign=0):
        """
        Space of modular symbols with given weight, character, and sign.

        INPUT:
            eps -- dirichlet.DirichletCharacter, the "Nebentypus" character.
            weight -- int, the weight >= 2
            sign -- int, either -1, 0, or 1
        EXAMPLES:
            sage: eps = DirichletGroup(4).gen(0)
            sage: eps.order()
            2
            sage: ModularSymbols(eps, 2)
            Full Modular Symbols space of level 4, weight 2, character [-1], sign 0, and dimension 0 over Cyclotomic Field of order 2 and degree 1
            sage: ModularSymbols(eps, 3)
            Full Modular Symbols space of level 4, weight 3, character [-1], sign 0, and dimension 2 over Cyclotomic Field of order 2 and degree 1

        We next create a space with character of order bigger than 2.
            sage: eps = DirichletGroup(5).gen(0)
            sage: eps     # has order 4
            [zeta_4]
            sage: ModularSymbols(eps, 2).dimension()
            0
            sage: ModularSymbols(eps, 3).dimension()
            2
        """            
        level = eps.modulus()
        ModularSymbolsAmbient.__init__(self, 
                weight = weight, 
                group = congroup.Gamma1(level),
                sign = sign, 
                base_field = eps.base_ring(),
                character = eps)
                
    def __repr__(self):
        return ("Full Modular Symbols space of level %s, weight %s, character %s, sign %s, " + \
                "and dimension %s over %s")%(self.level(), self.weight(), 
                    self.character(), self.sign(), self.dimension(), self.base_field())


    def _cuspidal_subspace_dimension_formula(self):
        if self.sign() == 0:
            m = 2
        else:
            m = 1
        return m * dims.dimension_cusp_forms_eps(self.character(), self.weight())
                    
    def _cuspidal_new_subspace_dimension_formula(self):
        if self.sign() == 0:
            m = 2
        else:
            m = 1
        return m * dims.dimension_new_cusp_forms(self.character(), self.weight())

    def _matrix_of_operator_on_modular_symbols(self, codomain, R, character_twist=False):
        """
        INPUT:
            self -- this space of modular symbols
            codomain -- space of modular symbols
            R -- list of lists [a,b,c,d] of length 4, which we view as elements of GL_2(Q).
            
        OUTPUT:
            a matrix, which represents the operator
            $$
               x \mapsto \sum_{g in R} g.x
            $$
            where g.x is the formal linear fractional transformation on modular symbols.
        """
        eps = self.character()
        rows = []
        for b in self.basis():
            v = formal_sum.FormalSum(0, check=False)
            for c, x in b.modular_symbol_rep():
                for g in R:
                    y = x.apply(g)
                    if character_twist:
                        v += y*c*eps(g[0])
                    else:
                        v += y*c
            w = codomain(v).vector()
            rows.append(w)
        M = matrix_space.MatrixSpace(self.base_field(), len(rows), codomain.degree(), sparse=False)
        return M(rows)

    def _degeneracy_matrix_raising(self, M):
        R = congroup.degeneracy_coset_representatives_gamma0(M.level(), self.level(), 1)
        return self._matrix_of_operator_on_modular_symbols(M, R, character_twist = True)

    def _dimension_formula(self):
        return None
 
    def boundary_space(self):
        try:
            return self.__boundary_space
        except AttributeError:
            pass
        self.__boundary_space = boundary.BoundarySymbols_wtk_eps(
            self.character(), self.weight(), self.sign())
        return self.__boundary_space

    def manin_symbols(self):
        try:
            return self.__manin_symbols
        except AttributeError:
            self.__manin_symbols = manin_symbols.ManinSymbolList_character(
                character=self.character(), weight=self.weight())
        return self.__manin_symbols

    def modular_symbols_of_level(self, N):
        """
        Returns a space of modular symbols with the same defining
        properties (weight, sign, etc.) as this space except with the
        level N.

        For example, if self is the space of modular symbols of weight
        2 for Gamma_0(22), and level is 11, then this function returns
        modular symbols of weight 2 for Gamma_0(11).
        """
        if self.level() % N == 0:
            eps = self.character().restrict(N)
        elif N % self.level() == 0:
            eps = self.character().extend(N)
        else:
            raise ValueError, "The level N (=%s) must be a divisor or multiple of the modulus of the character (=%s)"%(N, self.level())
        return ModularSymbolsAmbient_wtk_eps(eps, self.weight(), self.sign())

    def modular_symbols_of_sign(self, sign):
        """
        Returns a space of modular symbols with the same defining
        properties (weight, sign, etc.) as this space except with
        given sign.
        """
        sign = int(sign)
        if not (sign in [0,-1,1]):
            raise ValueError, "sign (=%s) must be an int in [0,-1,1]."%sign
        return ModularSymbolsAmbient_wtk_eps(self.character(), self.weight(), sign)

    def modular_symbols_of_weight(self, k):
        """
        Returns a space of modular symbols with the same defining
        properties (weight, sign, etc.) as this space except with
        weight k.
        """
        return ModularSymbolsAmbient_wtk_eps(self.character(), k, self.sign())
    

        
#########################################################################
## Automated testing code.
#########################################################################
        
def test_k2_g0_dim():
    """
    sage: from sage.all import *
    sage: ModularSymbols(11,sign=0).dimension()
    3
    sage: ModularSymbols(100,sign=0).dimension()
    31
    sage: ModularSymbols(22,sign=1).dimension()
    5
    """

def test_wtk_g0_dim():
    """
    sage: from sage.all import *
    sage: ModularSymbols(1,12).dimension()
    3
    sage: ModularSymbols(3,4).dimension()
    2
    sage: ModularSymbols(8,6,sign=-1).dimension()
    3
    """
    
if __name__ ==  '__main__':
    import doctest, sys
    from modsym import ModularSymbols
    from sage.modular.dirichlet import DirichletGroup
    doctest.testmod(sys.modules[__name__])

