"""
A single element of an ambient space of modular symbols.
"""

import sage.linalg.vector
import sage.misc.misc as misc
import sage.structure.formal_sum as formal_sum
import space

_print_mode = "manin"

def set_modsym_print_mode(mode="manin"):
    """
    Set the mode for printing of elements of modular symbols spaces.
    
    INPUT:
        mode -- a string.  The possibilities are as follows:
                * 'manin'   -- (the default) formal sums of Manin symbols [P(X,Y),(u,v)]
                * 'modular' -- formal sums of Modular symbols P(X,Y)*{alpha,beta},
                               where alpha and beta are cusps
                * 'vector'  -- as vectors on the basis for the ambient space
    """
    mode = str(mode).lower()
    if not (mode in ['manin', 'modular', 'vector']):
        raise ValueError, "mode must be one of 'manin', 'modular', or 'vector'"
    global _print_mode
    _print_mode = mode

class ModularSymbolElement:
    """
    A single element of a space of modular symbol.
    """
    def __init__(self, parent, vector):
        """
        INPUT:
            parent -- a space of modular symbols
            vector -- a vector that represents the modular symbol in terms
                      of a basis for the ambient space (not in terms of
                      a basis for parent!)
        """
        if not isinstance(parent, space.ModularSymbolsSpace):
            raise TypeError, "parent must be a ModularSymbolsSpace."
        if not isinstance(vector, sage.linalg.vector.Vector):
            raise TypeError, "vector must be of type Vector."
        if vector.degree() != parent.degree():
            raise TypeError, "vector (of degree %s) must be of degree the same as the degree of the parent (of degree %s)."%(vector.degree(), parent.degree())
        self.__parent = parent
        self.__vector = vector
        self.is_vector = True

    def __add__(self, right):
        if self.__parent != right.__parent:
            raise ArithmeticError, "parents must be the same"
        return ModularSymbolElement(self.__parent, self.__vector+right.__vector)
    
    def __cmp__(self, other):
        if not isinstance(other, ModularSymbolElement):
            if other is 0:
                if self.__vector == 0:
                    return 0
                else:
                    return -1
            other = self.__parent(other)
        if self.__parent != other.__parent:
            raise TypeError, "parents must be the same"
        print "cmp"
        return self.__vector.__cmp__(other.__vector)

    def __repr__(self):
        if _print_mode == "vector":
            return str(self.__vector)
        if _print_mode == "manin":
            m = self.manin_symbol_rep()
        elif _print_mode == "modular":
            m = self.modular_symbol_rep()
        c = [x[0] for x in m]
        v = [x[1] for x in m]
        return misc.repr_lincomb(v, c)

    def __mul__(self, right):
        return ModularSymbolElement(self.__parent, self.__vector*right)
    
    def __neg__(self):
        return ModularSymbolElement(self.__parent, -self.__vector)
    
    def __pos__(self):
        return self

    def __rmul__(self, left):
        return ModularSymbolElement(self.__parent, self.__vector*left)

    def __sub__(self, right):
        if self.__parent != right.__parent:
            raise ArithmeticError, "parents must be the same"
        return ModularSymbolElement(self.__parent, self.__vector - right.__vector)
    
    def coordinate_vector(self):
        if self.__parent.is_ambient():
            return self.__vector
        return self.parent().embedded_vector_space().coordinate_vector(self.__vector)

    def list(self):
        return self.__vector.list()

    def manin_symbol_rep(self):
        """
        Returns a representation of self as a formal sum of Manin symbols.

        (The result is cached for future use.)
        """
        try:
            return self.__manin_symbols
        except AttributeError:
            A = self.parent()
            v = self.vector()
            manin_symbols = A.ambient_space().manin_symbols_basis()
            ms = formal_sum.FormalSum([(v[i], manin_symbols[i]) for i in \
                  range(v.degree()) if v[i] != 0], check=False)
            self.__manin_symbols = ms
        return self.__manin_symbols
    
    def modular_symbol_rep(self):
        """
        Returns a representation of self as a formal sum of modular symbols.

        (The result is cached for future use.)
        """
        try:
            return self.__modular_symbols
        except AttributeError:
            A = self.parent()
            s = formal_sum.FormalSum([])
            for c, x in self.manin_symbol_rep():
                s += x.modular_symbol_rep()*c
            self.__modular_symbols = s
        return self.__modular_symbols

    def parent(self):
        return self.__parent

    def vector(self):
        return self.__vector

