"""nodoctest
Generic spaces of modular symbols.

All the spaces of modular symbols derive from this class.  This class
is an abstract base class.
"""

#*****************************************************************************
#       SAGE: System for Algebra and Geometry Experimentation
#
#       Copyright (C) 2005 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************


import math
import random
import weakref

import sage.linalg.vector_space as vector_space
import sage.linalg.matrix_space as matrix_space
import sage.misc.misc as misc
import sage.modular.dims as dims
import sage.modular.hecke as hecke
import sage.modular.modsym.element
import sage.rings.rings as rings
import sage.structure.gens as gens
import sage.structure.function as function
import sage.rings.arith as arith


class ModularSymbolsSpace(hecke.HeckeModule, gens.Gens):
    def __init__(self, group, weight, character, sign, base_field):
        self.__group = group
        self.__weight = weight
        self.__character = character
        self.__sign = sign
        self.__base_field = base_field
        hecke.HeckeModule.__init__(self)    
        
    def __cmp__(self, other):
        """
        Compare self and other.
        """
        if not isinstance(other, ModularSymbolsSpace):
            return -1
        if self.__group < other.__group:
            return -1
        elif self.__group > other.__group:
            return 1
        elif self.__weight < other.__weight:
            return -1
        elif self.__weight > other.__weight:
            return 1
        elif self.__character < other.__character:
            return -1
        elif self.__character > other.__character:
            return 1
        elif self.__sign < other.__sign:
            return -1
        elif self.__sign > other.__sign:
            return 1
        elif self.__base_field < other.__base_field:
            return -1
        elif self.__base_field > other.__base_field:
            return 1
        if self.is_ambient() and other.is_ambient():
            return 0
        if self.ambient_space() < other.ambient_space():
            return -1
        elif self.ambient_space() > other.ambient_space():
            return 1
        
        # ambients same and at most one of the two spaces is ambient
        if self.dimension() < other.dimension():
            return -1
        elif self.dimension() > other.dimension():
            return 1
            
        # both have same dimension and are in same ambient space, 
        # but neither is ambient
        if self.embedded_vector_space() == other.embedded_vector_space():
            return 0
        return -1

    def __contains__(self, x):
        if not isinstance(x, sage.modular.modsym.element.ModularSymbolElement):
            return False
        if x.parent() == self:  # easy
            return True
        if not self.ambient_space() == x.parent().ambient_space():
            return False
        if self.is_ambient():
            return True
        return x.vector() in self.embedded_vector_space()

    def _eigen_nonzero(self):
        try:
            return self.__eigen_nonzero
        except AttributeError:
            pass
        # Find the smallest integer i such that the i-th entries
        # of the entries of a basis for the dual vector space
        # are not all 0.  Then return the i-th basis vector for
        # the ambient space of modular symbols.
        A = self.ambient_space()
        V = self.embedded_dual_vector_space()
        B = V.basis()
        for i in range(A.dimension()):
            for b in B:
                if b[i] != 0:
                    self.__eigen_nonzero = i
                    return i
        assert False, 'bug in _eigen_nonzero'

    def _eigen_nonzero_element(self, n=1):
        r"""
        Return $T_n(x)$ where $x$ is a sparse modular symbol $x$ such that
        the image of $x$ is nonzero under the dual projection map
        associated to this space, and $T_n$ is the $n$-th Hecke operator.
        """
        if self.dimension() == 0:
            return self.ambient_space()(0)
        A = self.ambient_space()
        i = self._eigen_nonzero()
        return A._hecke_image_of_ith_basis_vector(n, i)

    def _modsym_eigenvalue(self, x):
        if not isinstance(x, sage.modular.modsym.element.ModularSymbolElement):
            raise TypeError, "x must be a modular symbol element."
        if not x in self.ambient_space():
            raise ArithmeticError, "x must be in the ambient space."
        v = self.dual_eigenvector()        
        return v.dot_product(x.vector())

    def ambient_space(self):
        raise NotImplementedError

    def base_field(self):
        return self.__base_field

    def base_ring(self):
        return self.__base_field
    
    def basis(self):
        """
        Returns a basis for self.
        """
        try:
            return self.__basis
        except:
            V = self.embedded_vector_space()
            self.__basis = [sage.modular.modsym.element.ModularSymbolElement(self, V.gen(i)) for \
                            i in range(V.dimension())]
        return self.__basis

    def character(self):
        return self.__character

    def complement(self):
        """
        Return the largest Hecke-stable subspace C of the ambient
        space that contains no simple factor isomorphic to a subspace
        of self.  In many cases, C is a complement to self, in the
        sense that the direct sum of C and self is the ambient space.
        """
        raise NotImplementedError

    def degree(self):
        return self.ambient_space().dimension()
                    
    def dimension(self):
        raise NotImplementedError

    def dual_eigenvector(self):
        """
        Return an eigenvector for the Hecke operators acting
        on this space via their transposes.  This eigenvector
        will have entries in an extension of the base ring
        of degree equal to the dimension of this space.

        INPUT:
            The input modular symbols space must Hecke equivariant
            and simple.
            
        OUTPUT:
            A vector with entries possibly in an extension of the base
            ring.  This vector is an eigenvector for all Hecke operators
            acting via their transpose.

        NOTES:
            (1) The answer is cached so subsequent calls always return
            the same vector.   However, the algorithm is randomized,
            so calls during another session may yield a different
            eigenvector.  This function is used mainly for computing
            systems of Hecke eigenvalues.

            (2) One can also view a dual eigenvector as defining (via
            dot product) a functional phi from the ambient space of
            modular symbols to a field.  This functional phi is an
            eigenvector for the dual action of Hecke operators on
            functionals. 
        """
        try:
            return self.__dual_eigenvector
        except AttributeError:
            pass
        
        # Find a Hecke operator that acts irreducibly on this space:
        p = 2
        t = self.dual_hecke_matrix(p)
        while True:
            f = t.charpoly()
            if f.is_simple():
                break
            p = arith.next_prime(p)
            t += random.choice([-2,-1,1,2]) * M.dual_hecke_operator(p)
        
        # Write down the eignvector.
        # Write f(x) = (x-alpha)*g(x), where alpha is a root
        # of f(x).
        n = f.degree()
        if n > 1:
            R = f.parent()
            K = R.quotient(f, 'alpha')    # Let K be the quotient R/(f),
                                          # with generator printed "alpha".
            alpha = K.gen()
            beta = ~alpha   # multiplicative inverse of alpha
            c = [-f[0]*beta]
            for i in range(1,n-1):
                c.append((c[i-1] - f[i])*beta)
            c.append( K(1) )
        else:
            K = self.base_ring()
            c = [1]

        # The entries of c are the coefficients of g (as stated in
        # William Stein's Ph.D. thesis, Section 3.5.3).  We compute
        # g(t)v for a some vector v, and get an eigenvector.
        V = vector_space.VectorSpace(K, n)   
        t = t.change_ring(K)      # coerce t to be over K.
        for j in range(n):
            v = V.gen(j)
            I = t.iterates(v, n)  # iterates v, v*t, v*t^2, ...
            w = V(0)
            for i in range(n):
                w += c[i]*V(I[i].list())
            if w != 0:
                break

        # Now w is an eigenvector for the action of the Hecke
        # operators on the subspace.  We need an eigenvector
        # in the original space, so we take the linear combination
        # of the basis for the embedded dual vector space given
        # by the entries of w.
        Vdual = self.embedded_dual_vector_space().change_ring(K)
        w_lift = Vdual.linear_combination_of_basis(w)

        # Finally rescale so the dot product of this vector and
        # the _eigen_nonzero_element is 1.
        alpha = w_lift.dot_product(self._eigen_nonzero_element().vector())
        w_lift = w_lift * (~alpha)
        
        self.__dual_eigenvector = w_lift
        return self.__dual_eigenvector

    def dual_hecke_matrix(self, n):
        """
        The matrix of the n-th Hecke operator acting on the
        embedded_dual_vector_space of self.
        """
        n = int(n)
        try:
            self._dual_hecke_matrices
        except:
            self._dual_hecke_matrices = {}
        if not self._dual_hecke_matrices.has_key(n):
            self._compute_dual_hecke_matrix(n)
        return self._dual_hecke_matrices[n] 

    def dual_star_involution_matrix(self):
        """
        Return the matrix of the dual star involution, which is
        induced by complex conjugation on the linear dual of modular
        symbols.
        """
        raise NotImplementedError
        
    def eigenvalue(self, n):
        n = int(n)
        try:
            return self.__eigenvalues[n]
        except AttributeError:
            self.__eigenvalues = {}
        except KeyError:
            pass
        if n <= 0:
            raise IndexError, "n must be a positive integer"

        if n == 1:
            a1 = self.base_ring()(1)
            self.__eigenvalues[1] = a1
            return a1

        if arith.is_prime(n):
            Tn_e = self._eigen_nonzero_element(n)
            an = self._modsym_eigenvalue(Tn_e)
            self.__eigenvalues[n] = an
            return an
        # Now use the Hecke eigenvalue recurrence, since arithmetic in
        # a field is faster than computing Heilbronn matrices for
        # non-prime n and doing some big sum (i.e., computing T_n(e)).
        # Also by computing using the recurrence on eigenvalues
        # we use information about divisors. 
        F = arith.factor(n)
        prod = None
        for p, r in F:
            (p, r) = (int(p), int(r))
            pow = p**r
            if not self.__eigenvalues.has_key(pow):
                # TODO: Optimization -- do something much more intelligent in case character is not defined.
                # For example, compute it using definition of <> operators on modular symbols!
                eps = self.character()
                if eps is None:
                    Tn_e = self._eigen_nonzero_element(pow)
                    self.__eigenvalues[pow] = self._modsym_eigenvalue(Tn_e)
                else:
                    # a_{p^r} := a_p * a_{p^{r-1}} - eps(p)p^{k-1} a_{p^{r-2}}
                    apr1 = self.eigenvalue(pow//p)
                    ap = self.eigenvalue(p)
                    k = self.weight()
                    apr2 = self.eigenvalue(pow//(p*p))
                    apow = ap*apr1 - eps(p)*(p**(k-1)) * apr2
                    self.__eigenvalues[pow] = apow
            if prod == None:
                prod = self.__eigenvalues[pow]
            else:
                prod *= self.__eigenvalues[pow]
        self.__eigenvalues[n] = prod
        return prod
        
    def embedded_vector_space(self):
        raise NotImplementedError

    def embedded_dual_vector_space(self):
        raise NotImplementedError

    def gen(self, n):
        """
        The n-th generator of self.
        
        INPUT:
            n -- int, an integer between 0 to the dimension minus 1.
            
        OUTPUT:
            -- The n-th basis element (ModularSymbol)
        """
        if n <0 or n>=self.dimension():
            raise IndexError
        return self.basis()[int(n)]

    def group(self):
        """
        Returns the group of this modular symbols space.

        INPUT:
           ModularSymbols self -- an arbitrary space of modular symbols
           
        OUTPUT:
           CongruenceSubgroup -- the congruence subgroup that this is a space
                              of modular symbols for.

        ALGORITHM:
           The group is recorded when this space is created.

        EXAMPLES:
        sage: m = ModularSymbols(20)
        sage: m.group()
        Gamma_0(20)
        """
        return self.__group

    def hecke_algebra(self, anemic=False):
        try:
            return self.__hecke_algebra
        except:
            self.__hecke_algebra = hecke.HeckeAlgebra(self, anemic)
            return self.__hecke_algebra

    def hecke_matrix(self, n):
        """
        The matrix of the n-th Hecke operator acting on the 
        basis for self.
        """
        n = int(n)
        if n <= 0:
            raise IndexError, "n must be positive."
        try:
            self._hecke_matrices
        except:
            self._hecke_matrices = {}
        if not self._hecke_matrices.has_key(n):
            self._compute_hecke_matrix(n)
        return self._hecke_matrices[n] 
    
    def hecke_operator(self, n):
        """
        Returns the n-th Hecke operator T_n.

        INPUT:
           ModularSymbols self -- Hecke equivariant space of
                                  modular symbols
           int n -- an integer at least 1.

        OUTPUT:
           function.MatrixFunction -- the matrix function that
                    represents the n-th Hecke operator on the chosen
                    basis for self.
        """
        n = int(n)
        #A = function.MatrixFunction(self.hecke_matrix(n), self, self, \
        #                            "Hecke operator T_%s on %s"%(n,self))
        return hecke.HeckeOperator(self.hecke_algebra(), None, n)

    def hecke_polynomial(self, n):
        """
        Return the characteristic polynomial of the n-th Hecke operator
        acting on this space.

        INPUT:
            n -- integer
        OUTPUT:
            a polynomial
        """
        n = int(n)
        try:
            return self.__hecke_polynomial[n]
        except AttributeError:
            self.__hecke_polynomial = {}
        except KeyError:
            pass
        f = self.hecke_matrix(n).charpoly()        
        self.__hecke_polynomial[n] = f 
        return f
        
    def is_ambient(self):
        raise NotImplementedError

    def is_anemic_hecke_module(self):
        raise NotImplementedError
        
    def is_full_hecke_module(self):
        raise NotImplementedError

    def is_new(self, p=None):
        raise NotImplementedError

    def is_cuspidal(self):
        raise NotImplementedError

    def is_simple(self):
        raise NotImplementedError
        
    def level(self):
        """
        Returns the level of this modular symbols space.

        INPUT:
           ModularSymbols self -- an arbitrary space of modular symbols
           
        OUTPUT:
           int -- the level

        ALGORITHM:
           The level is recorded when self is created.

        EXAMPLES:
        sage: m = ModularSymbols(20)
        sage: m.level()
        20
        """
        return self.__group.level()
        
    def ngens(self):
        """
        The number of generators of self.
        INPUT:
           ModularSymbols self -- arbitrary space of modular symbols.
        OUTPUT:
           int -- the number of generators, which is the same as the
                  dimension of self.
        ALGORITHM:
           Call the dimension function.
        EXAMPLES:
        sage: m = modsym.ModularSymbols(33)
        sage: m.ngens()
        9
        sage: m.dimension()
        9
        sage: modsym.ModularSymbols(100, weight=2, sign=1).ngens()
        18
        """
        return self.dimension()

    def projection(self):
        r"""
        Return the projection map from the ambient space to self.
        
        OUTPUT:
            MatrixFunction -- The output is a matrix function, which
                             defines projection from the ambient space
                             to self.

        ALGORITHM:
            Let $B$ be the matrix whose columns are got by
            concatenating together a basis for the factors of the
            ambient space.  Then the projection matrix onto self is
            the submatrix of $B^{-1}$ got from the rows corresponding
            to self, i.e., if the basis vectors for self appear as
            columns $n$ through $m$ of $B$, then the projection matrix
            is got from rows $n$ through $m$ of $B^{-1}$.  This is
            because projection with respect to the B basis is just
            given by an $m-n+1$ row slice $P$ of a diagonal matrix D
            with 1's in the $n$ through $m$ positions, so projection
            with respect to the standard basis is given by $P\cdot
            B^{-1}$, which is just rows $n$ through $m$ of $B^{-1}$.
        """

        # Compute the Hecke-stable projection map pi from the ambient
        # space of M to M.  Computing the projection map is the same
        # as writing the ambient space as a direct sum of M and its
        # Hecke-stable complement, which is the old subspace plus the
        # other new factors, then *inverting*.  With the projection
        # map in hand, we can compute Hecke operators directly on M
        # fairly quickly without having to compute them on the whole
        # ambient space.  Of course, computing this inverse is way too
        # much work to be useful in general (!).  (I sort of learned this
        # trick from Joe Wetherell, or at least he was aware of it
        # when I mentioned it to him in an airport once.  It's also
        # sort of like a trick Cremona uses in his book for elliptic
        # curves.)
        
        try:
            return self.__projection
        except:
            i = self.factor_number()
            if i == -1:
                raise NotImplementedError,\
                      "Computation of projection only implemented "+\
                      "for decomposition factors."
            A = self.ambient_space()
            B = A.decomposition_matrix_inverse()
            n = sum([A[j].dimension() for j in range(i)])
            C = B.submatrix_from_columns(range(n,n+self.dimension()))
            pi = function.MatrixFunction(C, A, self, \
                                         "Projection"%self)
            self.__projection = pi
            return self.__projection

    def q_expansion(self, prec):
        """
        Returns the q-expansion to precision prec of a newform
        associated to self, where self must be new, cuspidal, and
        simple.
        """
        try:
            f = self.__q_expansion
        except AttributeError:
            if not self.is_cuspidal():
                raise ArithmeticError, "self must be cuspidal."
            if not self.is_simple():
                raise ArithmeticError, "self must be simple."
            a2 = self.eigenvalue(2)
            R = rings.PowerSeriesRing(a2.parent(), "q")
            q = R.gen(0)
            f = q + a2*q**2 + rings.O(q**3)
        if f.prec() < prec:
            R = f.parent()
            ext = [self.eigenvalue(n) for n in range(f.prec(),prec)]
            f = R(list(f) + ext)
            self.__qeigenform = f.add_bigoh(prec)
            return self.__qeigenform
        else:
            return f.O(prec)

    def q_expansion_basis(self, prec):
        """
        Returns a basis of q-expansions (as power series) to precision
        prec of the space of modular forms associated to self.  The
        q-expansions are defined over the same base ring as prec.
        """
        try:
            return self.__q_expansion_basis
        except AttributeError:
            pass

        if not self.is_cuspidal():
            raise ArithmeticError, "space must be cuspidal"
        if self.is_simple():
            f = self.q_expansion(prec)
            R = PowerSeriesRing(self.base_ring(), 'q')
            B = [R([f[i][j] for i in xrange(prec)]) for j in range(self.dimension())]
            self.__q_expansion_basis = B
            return B
        B = []
        N = self.level()
        for A, _ in self.factorization():
            M = A.level()
            d = N//M
            C = A.q_expansion_basis(prec//d)
            assert N%M == 0
            if d == 1:
                B += C
            else:
                # replace q by q**d
                for f in C:
                    B.append(f.V(d))
        self.__q_expansion_basis = B
        return B
            
    def factorization(self):
        """
        Returns a list of pairs $(S,e)$ where $S$ is simple spaces of
        modular symbols and self is isomorphic to the direct sum of
        the $S^e$ as a module over the anemic Hecke algebra.

        ASSUMPTION: self is a module over the anemic Hecke algebra.
        """

        print "factorization is not done!!"
        try:
            self.__factorization
        except AttributeError:
            pass
        if self.is_simple():
            return [self]
        elif self.is_new():
            D = self.decomposition()
            if sign != 0:
                self.__factorization = D
            else:
                
                # take sign part of each factor
                
                raise NotImplementedError, "TODO"
            return D
        else:
            N = self.level()
            D = []
            d = 0
            A = self.ambient_space()
            for p in arith.prime_divisors(N):
                try:
                    M = A.modular_symbols_of_level(N//p)
                except ValueError:  # special case of character avoids certain levels
                    continue
                Eplus  = M.associated_new_simple_spaces(+1)
                Eminus = M.associated_new_simple_spaces(-1)
                if self.is_ambient():
                    D += E
                else:
                    # Have to laboriously check whether some image of each element
                    # of E lies in self.
                    
                    # TODO -- must take sum of +1 and -1 images of B!!
                    
                    1/0
                    for B in E:
                        im = A.subspace_generated_by_images(B)
                        inter = self.intersection(im)
                        if inter.dimension() > 0:
                            d += inter.dimension()
                            D.append(B)
                            if d >= self.dimension():  # done?
                                self.__factorization = D
                                return D
            self.__factorization = D
            return D

    def sign(self):
        """
        Returns the sign of self.

        For efficiency reasons, it is often useful to compute in the
        (largest) quotient of modular symbols where the * involution
        acts as +1, or where it acts as -1.
        

        INPUT:
           ModularSymbols self -- arbitrary space of modular symbols.
           
        OUTPUT:
           int -- the sign of self, either -1, 0, or 1.
                  -1 -- largest quotient where * acts as -1,
                  +1 -- largest quotient where * acts as +1, 
                   0 -- full space of modular symbols (no quotient).

        ALGORITHM:
           Call the dimension function.

        EXAMPLES:
        sage: m = ModularSymbols(33)
        sage: m.dimension()
        6
        sage: m.sign()
        1
        sage: m = ModularSymbols(33, sign=0)
        sage: m.sign()
        0
        sage: m.dimension()
        9
        sage: m = ModularSymbols(33, sign=-1)
        sage: m.sign()
        -1
        sage: m.dimension()
        3
        """
        return self.__sign

    def star_eigenvalue(self):
        """
        If the star involution acts as either the matrix +1 or -1 on
        self, this function returns +1 or -1, respectively.  Otherwise
        it returns 0.

        EXAMPLES:
            sage: M = ModularSymbols(11)
            sage: D = M.decomposition()
            sage: M.star_eigenvalue()
            0
            sage: D[0].star_eigenvalue()
            1
            sage: D[1].star_eigenvalue()
            0
            sage: D[1].plus_subspace().star_eigenvalue()
            1
            sage: D[1].minus_subspace().star_eigenvalue()
            -1
        """
        try:
            return self.__star_eigenvalue
        except AttributeError:
            pass
        if self.sign() != 0:
            return self.sign()
        M = self.star_involution().matrix()
        if M == 1:
            self.__star_eigenvalue = 1
            return 1
        elif M == -1:
            self.__star_eigenvalue = -1
            return -1
        self.__star_eigenvalue = 0
        return 0
    
    def star_decomposition(self):
        """
        """
        S = self.star_involution()
        return S.decomposition()

    def plus_subspace(self, compute_dual=True):
        """
        Return the subspace of self on which the star involution acts as +1.

        INPUT:
            compute_dual -- bool (default: True) also compute dual subspace.
                            This are useful for many algorithms.
                            
        OUTPUT:
            subspace of modular symbols
        """
        return self.sign_subspace(+1, compute_dual)
    
    def minus_subspace(self, compute_dual=True):
        """
        Return the subspace of self on which the star involution acts as -1.

        INPUT:
            compute_dual -- bool (default: True) also compute dual subspace.
                            This are useful for many algorithms.
        OUTPUT:
            subspace of modular symbols

        """
        return self.sign_subspace(-1, compute_dual)

    def sign_subspace(self, sign, compute_dual=True):
        """
        Return the subspace of self that is fixed under the star involution.

        INPUT:
            sign -- int (either -1, 0 or +1)
            compute_dual -- bool (default: True) also compute dual subspace.
                            This are useful for many algorithms.
        OUTPUT:
            subspace of modular symbols
        """
        sign = int(sign)
        if not sign in [-1, 0, 1]:
            raise ValueError, "sign must be -1, 0 or 1"
        if self.sign() == sign:  # an easy case
            return self
        if self.sign() == -sign:  # another easy case
            return self.zero_subspace()
        if sign == 0:
            # if sign is zero then self.sign() isn't 0 because
            # of the above checks.  
            raise ArithmeticError, "There is no sign 0 subspace of a space of modular symbols with nonzero sign."
        try:
            return self.__plus_subspace[(sign, compute_dual)]
        except AttributeError:
            self.__plus_subspace = {}
        except KeyError:
            pass
        P = self._compute_sign_subspace(sign, compute_dual)
        P.__star_eigenvalue = sign
        self.__plus_subspace[(sign,compute_dual)] = P
        return P
            
    def star_involution(self):
        """
        Return the star involution on self, which is induced by complex
        conjugation on modular symbols.
        """
        raise NotImplementedError

    def sturm_bound(self):
        r"""
        Returns the Sturm bound for this space of modular symbols.
        
        Returns a positive integer~$n$ such that the Hecke operators
        $T_1,\ldots, T_n$ acting on \emph{cusp forms} generate the
        Hecke algebra as a $\Z$-module when the character is trivial
        or quadratic.  Otherwise, $T_1,\ldots,T_n$ generate the Hecke
        algebra at least as a $\Z[\eps]$-module, where $\Z[\eps]$ is
        the ring generated by the values of the Dirichlet character
        $\eps$.  Alternatively, this is a bound such that if two cusp
        forms associated to this space of modular symbols are
        congruent modulo $(\lambda, q^n)$, then they are congruent
        modulo $\lambda$.
        
        REFERENCES:
        See the Agashe-Stein appendix to Lario and Schoof's \emph{Some
        computations with Hecke rings and deformation rings},
        Experimental Math., 11 (2002), no. 2, 303-311.  This result
        originated in the paper Sturm, \emph{On the congruence of
        modular forms}, Springer LNM 1240, 275--280, 1987.
        
        REMARK:

        Kevin Buzzard pointed out to me (William Stein) in Fall 2002
        that the above bound is fine for $\Gamma_1(N)$ with character,
        as one sees by taking a power of $f$.  More precisely, if $f
        \con 0 \pmod{p}$ for first $s$ coefficients, then $f^r \con 0
        \pmod{p}$ for first $sr$ coefficents.  Since the weight of
        $f^r$ is $r\cdot k(f)$, it follows that if $s \geq b$, where
        $b$ is the Sturm bound for $\Gamma_0(N)$ at weight $k(f)$, then
        $f^r$ has valuation large enough to be forced to be $0$ at
        $r*k(f)$ by Sturm bound (which is valid if we choose $r$
        correctly).  Thus $f \con 0 \pmod{p}$.  Conclusion: For
        $\Gamma_1(N)$ with fixed character, the Sturm bound is
        \emph{exactly} the same as for $\Gamma_0(N)$.

        A key point is that we are finding $\Z[\eps]$ generators for
        the Hecke algebra here, not $\Z$-generators.  So if one wants
        generators for the Hecke algebra over $\Z$, this bound must
        be suitably modified. 
        """
        # For Gamma_0(N), n = \frac{k}{12}[\SL_2(\Z):\Gamma_0(N)]
        try:
            return self.__sturm_bound
        except:
            self.__sturm_bound = \
                  int(math.ceil((self.weight() + dims.idxG0(self.level()))/12.0))
        return self.__sturm_bound

    def T(self, n):
        r"""
        Returns the $n$-th Hecke operator $T_n$.  This function is a
        synonym for \code{hecke_operator}.
        """
        return self.hecke_operator(n)
        
    def vector_space(self):
        """
        Returns the underlying vector space of self.
        NOTES:
        This is always an ambient vector space over the base ring of
        dimension equal to the dimension of self.  For the embedded
        subspace of the ambient space's vector space, use the
        embedded_subspace function.
        INPUT:
           ModularSymbols self -- arbitrary space of modular symbols.
        OUTPUT:
           VectorSpace
        ALGORITHM:
           Create and return the ambient vector space over the base
           field of dimension equal to the dimension of self.
        EXAMPLES:
        sage: import modsym
        sage: m = modsym.ModularSymbols(23)
        sage: m.vector_space()
        Full Vector space of degree 5 over Rational Field
        
        sage.: a = m[0]
        sgae.: a.vector_space()
        Full Vector space of degree 1 over Rational Field
        """

        try:
            return self.__vector_space
        except:
            self.__vector_space = vector_space.VectorSpace(
                        self.base_field(),self.dimension())
        return self.__vector_space


    def weight(self):
        """
        Returns the weight of this modular symbols space.

        INPUT:
           ModularSymbols self -- an arbitrary space of modular symbols
           
        OUTPUT:
           int -- the weight

        ALGORITHM:
           The weight is recorded when self is created.

        EXAMPLES:
        sage: m = ModularSymbols(20, weight=2)
        sage: m.weight()
        2
        """
        return self.__weight
    
    def zero_subspace(self):
        """
        Returns the zero subspace of this space of modular symbols.
        """
        return self.ambient_space()(self.vector_space().zero_subspace())




if __name__ ==  '__main__':
    import doctest, sys
    doctest.testmod(sys.modules[__name__])
