"""
Subspace of ambient spaces of modular symbols
"""

#*****************************************************************************
#       SAGE: System for Algebra and Geometry Experimentation
#
#       Copyright (C) 2005 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

import weakref

import ambient
import sage.linalg.vector_space as vector_space
import sage.misc.misc as misc
import sage.modular.congroup as congroup
import sage.rings.arith as arith
import sage.structure.function as function
import space

import sage.modular.modsym.element

class ModularSymbolsSubspace(space.ModularSymbolsSpace):
    """
    Subspace of ambient space of modular symbols
    """
    ################################
    # Special Methods
    ################################
    def __init__(self, ambient_space, vector_space):
        self.__ambient_space = ambient_space
        self.__vector_space = vector_space
        A = ambient_space
        space.ModularSymbolsSpace.__init__(self, A.group(), A.weight(), \
                                     A.character(), A.sign(), A.base_field())

    def __repr__(self):
        return "Dimension %s subspace of a modular symbols space of level %s"%(
                    self.dimension(), self.level())

    def __add__(self, other):
        if not isinstance(other, space.ModularSymbolsSpace):
            raise TypeError, "other (=%s) must be a ModularSymbolsSpace."%other
        if self.ambient_space() != other.ambient_space():
            raise ArithmeticError, "Sum only defined for subspaces of a " + \
                  "common ambient modular symbols space."
        if other.is_ambient():
            return other
        # Neither is ambient        
        V = self.embedded_vector_space() + other.embedded_vector_space()
        M = self.ambient_space()
        return M(V)

    def __call__(self, x, computed_with_hecke=False):
        """
        Coerce x into this modular symbols space (self). The result is
        either an element of self or a subspace of self.

        The allowed inputs are the same as for ambient spaces of
        modular symbols, except
        """
        if isinstance(x, sage.modular.modsym.element.ModularSymbolElement) and x.parent() == self:
            return x
        elif isinstance(x, vector_space.VectorSpace):
            return self._subspace(x, computed_with_hecke=computed_with_hecke)
        v = self.ambient_space()(x).vector()
        if not v in self.embedded_vector_space():
            raise ValueError, "v is not in the space of modular symbols"
        return sage.modular.modsym.element.ModularSymbolElement(self, v)
    

    ################################
    # Private functions
    ################################
    def _compute_hecke_matrix(self, n):
        A = self.ambient_space().hecke_matrix(n)
        if arith.gcd(self.level(), n) != 1:
            try:
                check = not self._is_full_hecke_module
            except AttributeError:
                check = True
        else:
            try:
                check = not self._is_anemic_hecke_module
            except AttributeError:
                check = True
            
        T = A.restrict(self.embedded_vector_space(), check=check)
        self._hecke_matrices[n] = T
        return T

    def _compute_dual_hecke_matrix(self, n):
        A = self.ambient_space().dual_hecke_matrix(n)
        if arith.gcd(self.level(), n) != 1:
            try:
                check = not self._is_full_hecke_module
            except AttributeError:
                check = True
        else:
            try:
                check = not self._is_anemic_hecke_module
            except AttributeError:
                check = True
                
        T = A.restrict(self.embedded_dual_vector_space(), check=check)
        self._dual_hecke_matrices[n] = T
        return T

    def _set_dual_vector_space(self, V):
        self.__dual_vector_space = V

    ################################
    # Public functions
    ################################
    def ambient_space(self):
        return self.__ambient_space

    def ambient_vector(self, v):
        """
        Given a vector v in self.vector_space(), return the
        corresponding element in the vector space of the ambient
        modular symbols space.
        """
        B = self.__vector_space.basis()
        zero = self.__vector_space.ambient_space()(0)
        return sum([v[i]*B[i] for i in range(len(B))], zero)

    def atkin_lehner_operator(self, d):
        """
        Return the Atkin-Lehner operator W_d on this space, where d is
        a divisor of the level N such that N/d and d are coprime.
        """
        d = int(d)
        if self.level() % d != 0:
            raise ArithmeticError, "d (=%s) must be a divisor of the level (=%s)"%(d,self.level())

        N = self.level()
        for p, e in arith.factor(d):
            v = arith.valuation(N, p)
            if e < v:
                d *= p**(v-e)
        d = int(d)

        try:
            return self.__atkin_lehner_operator[d]
        except AttributeError:
            self.__atkin_lehner_operator = {}
        except KeyError:
            pass
        W = self.ambient_space().atkin_lehner_operator(d).restrict(self)
        W.name("Atkin-Lehner operator W_%s"%d)
        self.__atkin_lehner_operator[d] = W
        return W

    def boundary_map(self):
        """
        The boundary map to the corresponding space of boundary
        modular symbols.  (This is the restriction of the map on the
        ambient space.)
        """
        try:
            return self.__boundary_map
        except AttributeError:
            # restrict from ambient space
            b = self.ambient_space().boundary_map()
            self.__boundary_map = b.restrict_domain(self)
            return self.__boundary_map

    def complement(self):
        """
        Return the largest Hecke-stable complement of this space.
        """
        if self.is_ambient():
            return self.ambient_space().zero_subspace()

        if self.is_full_hecke_module(compute=False):
            anemic = False
        elif self.is_anemic_hecke_module():
            anemic = True
        else:
            raise ArithmeticError, "dual subspace only defined for Hecke (or, in some cases, anemic Hecke) modules."

        # TODO: optimize in some cases by computing image of complementary factor instead of kernel...?
        
        misc.verbose("computing")
        N = self.level()
        A = self.ambient_space()
        V = A.vector_space()
        p = 2
        while True:
            misc.verbose("using T_%s"%p)
            if anemic:
                while N % p == 0: p = arith.next_prime(p)
            f = self.hecke_polynomial(p)
            T = A.hecke_matrix(p)
            g = T.charpoly()
            V = T.kernel_on(V, poly=g//f, check=False)
            if V.dimension() + self.dimension() <= A.dimension():
                break
            p = arith.next_prime(p)
            if p > A.sturm_bound() and p > A.hecke_bound():  # to avoid computing hecke bound unless necessary
                break

        if V.dimension() + self.dimension() == A.dimension():
            return A(V)
        else:
            # failed
            raise RuntimeError, "Computation of complementary space failed (cut down to dimension %s, but should have cut down to dimension %s)."%(V.dimension(), self.dimension())


    def cuspidal_subspace(self):
        """
        Return the cuspidal subspace of this space of modular symbols.
        """
        try:
            return self.__cuspidal_subspace
        except AttributeError:
            S = self.ambient_space().cuspidal_subspace()            
            self.__cuspidal_subspace = S.intersection(self)
            return self.__cuspidal_subspace

    def degeneracy_map(self, level, t=1):
        """
        The t-th degeneracy map from self to the space of ambient
        modular symbols of the given level.  The level of self must be
        a divisor or multiple of level, and t must be a divisor of the
        quotient.
        
        INPUT:
            level -- int, the level of the codomain of the map (positive int).
            t  -- int, the parameter of the degeneracy map, i.e., the map is
                  related to f(q) |--> f(q^t).
                  
        OUTPUT:
            A linear function from self to the space of modular symbols
            of given level with the same weight, character, sign,
            etc., as this space.
        """
        d = self.ambient_space().degeneracy_map(level, t)
        d2 = d.matrix().restrict_domain(self.embedded_vector_space())
        return function.MatrixFunction(self, d.codomain(), d2, d.name())

    def dimension(self):
        return self.__vector_space.dimension()

    def dual_star_involution_matrix(self):
        """
        Return the matrix of the dual star involution, which is
        induced by complex conjugation on the linear dual of modular
        symbols.
        """
        try:
            return self.__dual_star_involution
        except AttributeError:
            pass
        S = self.ambient_space().dual_star_involution_matrix()
        A = S.restrict(self.embedded_dual_vector_space())
        self.__dual_star_involution = A
        return self.__dual_star_involution

    def eisenstein_subspace(self):
        """
        Return the Eisenstein subspace of this space of modular symbols.
        """
        try:
            return self.__eisenstein_subspace
        except AttributeError:
            S = self.ambient_space().eisenstein_subspace()            
            self.__eisenstein_subspace = S.intersection(self)
            return self.__eisenstein_subspace


    def embedded_vector_space(self):
        return self.__vector_space

    def __compute_embedded_dual_space_from_subspace(self):
        """
        ALGORITHM: Compute the char poly of each Hecke operator on the
        subspace, then use it to cut out a subspace of the dual.  If
        the dimension cuts down to the dimension of self terminate
        with success.  If it stays bigger beyond the bound (Sturm)
        bound, raise a RuntimeError exception.
        """
        if self.is_ambient():
            return self.vector_space()

        if self.is_full_hecke_module(compute=False):
            anemic = False
        elif self.is_anemic_hecke_module():
            anemic = True
        else:
            raise ArithmeticError, "dual subspace only defined for Hecke (or, in some cases, anemic Hecke) modules."

        # TODO: optimize in some cases by computing image of complementary factor instead of kernel...?
        
        misc.verbose("computing")
        N = self.level()
        A = self.ambient_space()
        V = A.vector_space()
        p = 2
        while True:
            misc.verbose("using T_%s"%p)
            if anemic:
                while N % p == 0: p = arith.next_prime(p)
            f = self.hecke_polynomial(p)
            T = A.dual_hecke_matrix(p)
            V = T.kernel_on(V, poly=f, check=False)
            if V.dimension() <= self.dimension():
                break
            p = arith.next_prime(p)
            if p > A.sturm_bound() and p > A.hecke_bound():  # to avoid computing hecke bound unless necessary
                break

        if V.dimension() == self.dimension():
            return V
        else:
            # failed
            raise RuntimeError, "Computation of embedded dual vector space failed (cut down to dimension %s, but should have cut down to dimension %s)."%(V.dimension(), self.dimension())

    def hecke_bound(self):
        if self.is_cuspidal():
            return self.sturm_bound()
        else:
            return self.ambient_space().hecke_bound()
    
    def embedded_dual_vector_space(self):
        """
        Compute embedded dual vector space if possible.  In general
        this won't be possible, e.g., if this space is not Hecke
        equivariant, possibly if it is not cuspidal, or if the
        characteristic is not 0.  In all these cases we raise a
        RuntimeError exception.
        """
        try:
            return self.__dual_vector_space
        except AttributeError:
            self.__dual_vector_space = self.__compute_embedded_dual_space_from_subspace()
        return self.__dual_vector_space                                   
        
    def intersection(self, other):
        """
        Returns the intersection of self and other, which must both
        lie in a common ambient space of modular symbols.
        EXAMPLES:
            sage: M = ModularSymbols(43, sign=1)
            sage: A = M[0] + M[1]
            sage: B = M[1] + M[2]
            sage: A.dimension(), B.dimension()
            (2, 3)
            
            sage.: C = A.intersection(B); C.dimension()
            1
        """
        if self.ambient_space() != other.ambient_space():
            raise ArithmeticError, "Intersection only defined for subspaces of"\
                  + " a common ambient modular symbols space."
        if other.is_ambient():
            return self
        # Neither is ambient
        V = self.embedded_vector_space().intersection(other.embedded_vector_space())
        M = self.ambient_space()
        A = M(V)
        if self.is_full_hecke_module(compute=False) and other.is_full_hecke_module(compute=False):
            A._is_full_hecke_module = True
            A._is_anemic_hecke_module = True
        elif self.is_anemic_hecke_module() and other.is_anemic_hecke_module():
            A._is_anemic_hecke_module = True
        return A

    def is_ambient(self):
        return self.dimension() == self.ambient_space().dimension()

    def is_anemic_hecke_module(self):
        """
        If compute is True (the default), return True if this space is
        known to be invariant (because of how it was computed) under
        all Hecke operators of index coprime to the level.  If this
        function returns False, the space still might be an anemic
        Hecke module, but we have no way of telling.
        """
        try:
            return self._is_anemic_hecke_module
        except AttributeError:
            if self.is_full_hecke_module(False):
                self._is_anemic_hecke_module = True
                return True
            else:
                return False

    def is_full_hecke_module(self, compute=True):
        """
        Return True if this space is invariant under all Hecke
        operators.  If compute = False (not the default), only return
        True if the space is known to be invariant under all Hecke
        operators because of how it was computed.
        """
        try:
            return self._is_full_hecke_module
        except AttributeError:
            pass

        try:
            if not self._is_anemic_hecke_module:
                self._is_full_hecke_module = False
                return False
        except AttributeError:
            pass

        if not compute:
            return False
        
        # now have to compute whether invariant under Hecke
        # operators of index dividing the level
        misc.verbose("Determining if Hecke module is full.")
        N = self.level()
        B = self.hecke_bound()
        if self.is_anemic_hecke_module():
            P = arith.prime_divisors(N)
        else:
            P = arith.prime_range(B)
        for p in P:
            if not self.is_hecke_invariant(p):
                self._is_full_hecke_module = False
                if N % p != 0:
                    self._is_anemic_hecke_module = False
                return False
        self._is_full_hecke_module = True
        return True

    def is_hecke_invariant(self, n=None):
        """
        Return True if this space is invariant under the
        n-th Hecke operator.

        INPUT:
            n -- an integer, or None which means "all Hecke operators"
        OUTPUT:
            bool
        """
        #print "skipping known in subspace.py: is_hecke_invariant"; n = int(n)
        #if False:
        if n is None:
            return self.is_full_hecke_module()
        n = int(n)
        N = self.level()
        if arith.gcd(N,n) == 1:
            if self.is_anemic_hecke_module():
                return True
        else:
            try:
                if self._is_full_hecke_module:
                    return True
            except AttributeError:
                pass

        # Have to compute
        A = self.ambient_space()
        T = A.hecke_matrix(n)
        V = self.embedded_vector_space()
        try:
            T.restrict(V, check=True)
        except ArithmeticError:
            return False
        else:
            return True
        

    def is_new(self, p=None):
        try:
            if self.__is_new.has_key(p):
                return self.__is_new[p]
        except AttributeError:
            N = self.ambient_space().new_subspace(p).embedded_vector_space()
            self.__is_new[p] = self.embedded_vector_space().is_subspace(N)
            return self.__is_new[p]

    def is_cuspidal(self):
        try:
            return self.__is_cuspidal
        except AttributeError:
            C = self.ambient_space().cuspidal_subspace()
            self.__is_cuspidal = self.is_subspace(C)
            return self.__is_cuspidal

    def is_eisenstein(self):
        try:
            return self.__is_eisenstien
        except AttributeError:
            C = self.ambient_space().eisenstein_subspace()
            self.__is_eisenstein = self.is_subspace(C)
            return self.__is_eisenstein

    def is_subspace(self, V):
        """
        Returns True if and only if self is a subspace of V.
        """
        return self.ambient_space() == V.ambient_space() and \
               self.embedded_vector_space().is_subspace(V.embedded_vector_space())

    def linear_combination_of_basis(self, v):
        return self(self.embedded_vector_space().linear_combination_of_basis(v))

    def new_subspace(self, p=None):
        """
        Return the cuspidal subspace of this space of modular symbols.
        """
        try:
            if self.__is_new[p]:
                return self
        except AttributeError:
            self.__is_new = {}
        except KeyError:
            pass

        if self.dimension() == 0:
            self.__is_new[p] = True
            return self
        try:
            return self.__new_subspace[p]
        except AttributeError:
            self.__new_subspace = {}
        except KeyError:
            pass

        S = self.ambient_space().new_subspace(p)
        ns = S.intersection(self)
        if ns.dimension() == self.dimension():
            self.__is_new[p] = True
        ns.__is_new = {p:True}
        self.__new_subspace[p] = ns
        return self.__new_subspace[p]

    def _compute_sign_subspace(self, sign, compute_dual=True):
        """
        Return the subspace of self that is fixed under the star involution.

        INPUT:
            sign -- int (either -1 or +1)
            compute_dual -- bool (default: True) also compute dual subspace.
                            This are useful for many algorithms.
        OUTPUT:
            subspace of modular symbols

        """
        S = self.star_involution().matrix() - sign
        V = S.kernel()
        if compute_dual:
            Sdual = self.dual_star_involution_matrix() - sign
            Vdual = Sdual.kernel()
            return self._subspace(V, Vdual)
        else:
            return self._subspace(V)

    def _subspace(self, V, Vdual=None, computed_with_hecke=False):
        """
        INPUT:
            V -- subspace of self.vector_space(); so the degree of V is the same
                 as the *dimension* of self
            Vdual -- subspace of dual vector space of self
        OUTPUT:
            modular symbols subspace of self
        """
        E = self.embedded_vector_space()
        M_V = V.matrix()
        M_E = E.matrix()
        # We encode the operation of taking the linear combinations of
        # the basis of E given by the basis of V as a single matrix
        # multiplication, since matrix multiplication is (presumed to be)
        # so fast, and their are asymptotically fast algorithms. 
        A = M_V * M_E
        space = self.ambient_space()(A.row_space())
        if not (Vdual is None):
            E = self.embedded_dual_vector_space()
            M_Vdual = Vdual.matrix()
            M_E = E.matrix()
            A = M_Vdual * M_E
            space._set_dual_vector_space(A.row_space())
        if computed_with_hecke:
            if self.is_full_hecke_module(compute=False):
                space._is_full_hecke_module = True
            elif self.is_anemic_hecke_module():
                space._is_anemic_hecke_module = True
        return space

    def star_involution(self):
        """
        Return the star involution on self, which is induced by complex
        conjugation on modular symbols.
        """
        try:
            return self.__star_involution
        except AttributeError:
            pass
        S = self.ambient_space().star_involution().matrix()
        A = S.restrict(self.embedded_vector_space())
        self.__star_involution = function.MatrixFunction(A, self, self, "Star involution on %s"%self)
        return self.__star_involution

if __name__ ==  '__main__':
    import doctest, sys
    from modsym import ModularSymbols
    from sage.modular.dirichlet import DirichletGroup
    doctest.testmod(sys.modules[__name__])
