"""
MPFR interface

AUTHORS: Kyle Schalm <kschalm@math.utexas.edu> and William Stein
"""

#*****************************************************************************
#
#   SAGE: System for Algebra and Geometry Experimentation    
#
#       Copyright (C) 2005 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************


include 'interrupt.pxi'

cimport _ring
import _ring
cimport _element
import _element
import sage.rings.coerce
import operator


#*****************************************************************************
# Headers.  When you past things in here from mpfr, be sure
# to remove const's, since those aren't allowed in pyrex.  Also, it can be
# challenging figuring out how to modify things from mpfr.h to be valid pyrex
# code.    Note that what is here is only used for generating the C code.
# The C compiler doesn't see any of this -- it only sees mpfr.h and stdlib.h
#*****************************************************************************

cdef extern from "stdlib.h":
    ctypedef int size_t 
    void free(void *ptr)

cdef extern from "mpfr.h":
    ctypedef struct __mpfr_struct:
        pass
    ctypedef __mpfr_struct mpfr_t[1]
    ctypedef enum mpfr_rnd_t:
        GMP_RNDN = 0
        GMP_RNDZ = 1
        GMP_RNDU = 2
        GMP_RNDD = 3
        GMP_RND_MAX = 4
        GMP_RNDNA = -1
    
    ctypedef mpfr_rnd_t mp_rnd_t
    ctypedef long int mp_exp_t
    ctypedef long mp_prec_t

    #mp_rnd_t GMP_RNDZ, GMP_RNDN, GMP_RNDU, GMP_RNDD

    void mpfr_init (mpfr_t x)
    void mpfr_init2 (mpfr_t x, mp_prec_t prec)
    void mpfr_clear (mpfr_t x)

    int mpfr_set (mpfr_t rop, mpfr_t op, mp_rnd_t rnd)
    int mpfr_set_str (mpfr_t rop, char *s, int base, mp_rnd_t rnd)

    char * mpfr_get_str (char *str, mp_exp_t *expptr, int base, size_t n, mpfr_t op, mp_rnd_t rnd)

    # Arithmetic
    int mpfr_add (mpfr_t rop, mpfr_t op1, mpfr_t op2, mp_rnd_t rnd)
    int mpfr_sub (mpfr_t rop, mpfr_t op1, mpfr_t op2, mp_rnd_t rnd)
    int mpfr_mul (mpfr_t rop, mpfr_t op1, mpfr_t op2, mp_rnd_t rnd)
    int mpfr_div (mpfr_t rop, mpfr_t op1, mpfr_t op2, mp_rnd_t rnd)

    # Special functions
    int mpfr_log (mpfr_t rop, mpfr_t op, mp_rnd_t rnd)
    int mpfr_log2 (mpfr_t rop, mpfr_t op, mp_rnd_t rnd)
    int mpfr_log10 (mpfr_t rop, mpfr_t op, mp_rnd_t rnd)

    int mpfr_exp (mpfr_t rop, mpfr_t op, mp_rnd_t rnd)
    int mpfr_exp2 (mpfr_t rop, mpfr_t op, mp_rnd_t rnd)
    int mpfr_exp10 (mpfr_t rop, mpfr_t op, mp_rnd_t rnd)

    int mpfr_cos (mpfr_t rop, mpfr_t op, mp_rnd_t rnd)
    int mpfr_sin (mpfr_t rop, mpfr_t op, mp_rnd_t rnd)
    int mpfr_tan (mpfr_t rop, mpfr_t op, mp_rnd_t rnd)    

#*****************************************************************************
#
#             Implementation
#
#*****************************************************************************



#*****************************************************************************
#
#       Real Field
#
#*****************************************************************************
# I'm putting the real field in Pyrex, so mpfr elements will have
# access to their parent via direct C calls.  This will be faster.

cdef class RealField(_ring.Ring):
    cdef int prec
    cdef mp_rnd_t rnd

    def __init__(self, int prec=64, int rnd=GMP_RNDN):
        """
        RealField(prec, rnd):

        INPUT:
            prec -- (int) precision (must be >= 2)
            rnd -- direction to round:   TODO: these should be strings 
                      0 -- RNDN
                      1 -- RNDZ
                      2 -- RNDU
                      3 -- RNDD
        OUTPUT:
            RealField -- a real field implemented using MPFR.
            
        EXAMPLES:
            sage: RealField(10)
            Real Field of precision 10
            sage: RealField()
            Real Field of precision 64
            sage: RealField(100000)
            Real Field of precision 100000
        """
        if prec < 2:
            raise ValueError, "prec (=%s) must be >= 2."%prec
        self.prec = prec
        self.rnd = rnd

    def __repr__(self):
        s = "Real Field of precision %s"%self.prec
        if self.rnd != GMP_RNDN:
            s = s + " and rounding %s"%(self.rnd)
        return s

    def __call__(self, x):
        """
        Coerce x into this real field.

        EXAMPLES:
            sage: R = RealField(10)
            sage: R('1.234')
            0.1232e1
        """
        return RealNumber(self, x)

    def __contains__(self, x):
        """
        True if x is in this field.
        
        EXAMPLES:
            sage: R = RealField(10)
            sage: 1 in R
            False
            sage: R(1.0) in R
            True
        """
        return isinstance(x, RealNumber) and x.parent() == self

    def __cmp__(self, other):
        """
        EXAMPLES:
            sage: RealField(10) == RealField(11)
            False
            sage: RealField(10) == RealField(10)
            True
            sage: RealField(10,0) == RealField(10,1)
            False
            sage: RealField(10) == IntegerRing()
            False
        """
        if not isinstance(other, RealField):
            return -1
        cdef RealField _other
        _other = other  # to access C structure
        if self.prec == _other.prec and self.rnd == _other.rnd:
            return 0
        return 1

    def is_atomic_repr(self):
        """
        Returns True, to signify that elements of this field print
        without sums, so parenthesis aren't required, e.g., in
        coefficients of polynomials.

        EXAMPLES:
            sage: RealField(10).is_atomic_repr()
            True
        """
        return True

    def is_field(self):
        """
        Returns True, to signify that this is a field.

        EXAMPLES:
            sage: RealField(10).is_field()
            True
        """
        return True

    def is_finite(self):
        """
        Returns False, since the field of real numbers is not finite.
        
        EXAMPLES:
            sage: RealField(10).is_finite()
            False
        """
        return False

    def characteristic(self):
        """
        Returns 0, since the field of real numbers has characteristic 0.

        EXAMPLES:
            sage: RealField(10).characteristic()
            0
        """
        return 0
    
    def name(self):
        return "RealField%s_%s"%(self.prec,self.rnd)
    
    def __hash__(self):
        return hash(self.name())
    

R = RealField()

#*****************************************************************************
#
#     RealNumber -- element of Real Field
#
#*****************************************************************************
cdef class RealNumber(_element.Element):
    cdef mpfr_t value
    cdef RealField _parent

    def __new__(self, RealField parent, x=None):
        self._parent = parent
        mpfr_init2(self.value, parent.prec)

    # Important WARNING: Do *not* create RealNumbers directly!
    # Instead use the call method of a RealField.  Otherwise if
    # you neglect to give a RealField as first argument then
    # SAGE crashes for (heretofor) unknown reasons. 
    def __init__(self, RealField parent, x=None):
        """
        Create a real number.  Should be called by first creating
        a RealField, as illustrated in the examples.

        EXAMPLES:
            sage: R = RealField()
            sage: R('1.2456')
            0.12455999999999999999e1
            sage: R = RealField(3)
            sage: R('1.2456')
            0.10e1
        """
        if x is None:
            return
        cdef RealNumber _x
        if isinstance(x, RealNumber):
            _x = x  # so we can get at x.value
            mpfr_set(self.value, _x.value, GMP_RNDZ)
        # ...
        # add many more cases  -- be careful to use base 16 or something...
        # ...
        else:
            s = str(x)
            if mpfr_set_str(self.value, s, 10, GMP_RNDZ):
                raise TypeError, "Unable to convert x (='%s') to mpfr."%s

    def  __dealloc__(self):
        if self.value:
            mpfr_clear(self.value)

    def __repr__(self):
        return self.str()

    def parent(self):
        """
        EXAMPLES:
            sage: R = RealField()
            sage: a = R('1.2456')
            sage: a.parent()
            Real Field of precision 64
        """
        return self._parent

    def str(self, int base=10):
        cdef char *s
        cdef mp_exp_t exponent

        _sig_on
        s = mpfr_get_str(<char*>0, &exponent, base, 0, self.value, GMP_RNDZ)
        _sig_off
        if s == <char*> 0:
            raise RuntimeError, "Unable to convert an mpfr number to a string."
        t = str(s)
        free(s)
        if len(t) > 0 and t[0] == "-":
            return "-0.%se%s"%(t[1:], exponent)
        return "0.%se%s"%(t, exponent)

    def __add(RealNumber self, RealNumber other):
        cdef RealNumber x
        x = RealNumber(self._parent)
        mpfr_add(x.value, self.value, other.value, self._parent.rnd)
        return x
        
    def __add__(x, y):
        """
        EXAMPLES:
            sage: R = RealField()
            sage: R(-1.5) + R(2.5)
            0.10000000000000000000e1
        """
        if isinstance(x, RealNumber) and isinstance(y, RealNumber):
            return x.__add(y)
        return sage.rings.coerce.bin_op(x, y, operator.add)
    
    def __sub(RealNumber self, RealNumber other):
        cdef RealNumber x
        x = RealNumber(self._parent)
        mpfr_sub(x.value, self.value, other.value, self._parent.rnd)
        return x
        
    def __sub__(x, y):
        """
        EXAMPLES:
            sage: R = RealField()
            sage: R(-1.5) - R(2.5)
            -0.40000000000000000000e1
        """
        if isinstance(x, RealNumber) and isinstance(y, RealNumber):
            return x.__sub(y)
        return sage.rings.coerce.bin_op(x, y, operator.sub)
        
    def __mul(RealNumber self, RealNumber other):
        cdef RealNumber x
        x = RealNumber(self._parent)
        mpfr_mul(x.value, self.value, other.value, self._parent.rnd)
        return x
    
    def __mul__(x, y):
        """
        EXAMPLES:
            sage: R = RealField()
            sage: R(-1.5) * R(2.5)
            -0.37500000000000000000e1
        """
        if isinstance(x, RealNumber) and isinstance(y, RealNumber):
            return x.__mul(y)
        return sage.rings.coerce.bin_op(x, y, operator.mul)
    
    def __div(RealNumber self, RealNumber other):
        if not other:
            raise ZeroDivisionError, "RealNumber division by zero"
        cdef RealNumber x
        x = RealNumber(self._parent)
        mpfr_div(x.value, self.value, other.value, self._parent.rnd)
        return x
        
    def __div__(x, y):
        """
        EXAMPLES:
            sage: R = RealField()
            sage: R(-1.5) / R(2.5)
            -0.60000000000000000002e0
        """
        if isinstance(x, RealNumber) and isinstance(y, RealNumber):
            return x.__div(y)
        return sage.rings.coerce.bin_op(x, y, operator.div)
    


    ############################
    # Special Functions
    ############################
    def log(self):
        """
        EXAMPLES:
            sage: R = RealField()
            sage: R(2).log()
            0.69314718055994530942e0
        """
        cdef RealNumber x
        x = RealNumber(self._parent)
        mpfr_log(x.value, self.value, self._parent.rnd)
        return x

    def log2(self):
        cdef RealNumber x
        x = RealNumber(self._parent)
        mpfr_log2(x.value, self.value, self._parent.rnd)
        return x

    def log10(self):
        cdef RealNumber x
        x = RealNumber(self._parent)
        mpfr_log10(x.value, self.value, self._parent.rnd)
        return x

    def exp(self):
        cdef RealNumber x
        x = RealNumber(self._parent)
        mpfr_exp(x.value, self.value, self._parent.rnd)
        return x

    def exp2(self):
        cdef RealNumber x
        x = RealNumber(self._parent)
        mpfr_exp2(x.value, self.value, self._parent.rnd)
        return x

    def exp10(self):
        cdef RealNumber x
        x = RealNumber(self._parent)
        mpfr_exp10(x.value, self.value, self._parent.rnd)
        return x

    def cos(self):
        cdef RealNumber x
        x = RealNumber(self._parent)
        mpfr_cos(x.value, self.value, self._parent.rnd)
        return x
    
    def sin(self):
        """
        EXAMPLES:
            sage: R = RealField(100)
            sage: R(2).sin()
            0.9092974268256816953960198659115e0
        """
        cdef RealNumber x
        x = RealNumber(self._parent)
        mpfr_sin(x.value, self.value, self._parent.rnd)
        return x
    
    def tan(self):
        cdef RealNumber x
        x = RealNumber(self._parent)
        mpfr_tan(x.value, self.value, self._parent.rnd)
        return x
