"""
Quotients of univariate polynomial rings
"""

#*****************************************************************************
#       Copyright (C) 2005 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

import operator

import coerce
import element
import number_field
import number_field_element
import polynomial
import polynomial_ring
import rational_field
import ring

import sage.misc.misc as misc
import sage.rings.arith as arith
import sage.structure.gens as gens


####################################################################
# Rings
####################################################################

class PolynomialQuotientRing(ring.Ring, gens.Gens):
    """
    Quotient of univariate polynomial ring.
    """
    def __init__(self, R, f, variable=None):
        r"""
        Create a quotient of a polynomial ring.
        
        INPUT:
            R -- polynomial ring in one variable.
            f -- element of R.
        OUTPUT:
            Creates the quotient ring R/(f).

        EXAMPLES:
        
        We create the quotient ring $\Z[x]/(x^3+7)$, and demonstrate many
        basic functions with it:
        
            sage: Z = IntegerRing()
            sage: R = PolynomialRing(Z,'x'); x = R.gen()
            sage: S = R.quotient(x**3 + 7, 'a'); a = S.gen()
            sage: S
            Univariate Quotient Polynomial Ring in a over Integer Ring with modulus x^3 + 7
            sage: a**3
            -7
            sage: S.is_field()
            False
            sage: a in S
            True
            sage: x in S
            False
            sage: S.polynomial_ring()
            Univariate Polynomial Ring in x over Integer Ring
            sage: S.modulus()
            x^3 + 7
            sage: S.degree()
            3

        We create the ``iterated'' polynomial ring quotient
        $$
               R = (\F_2[y]/(y^{2}+y+1))[x]/(x^3 - 5).
        $$
        
            sage: A = PolynomialRing(GF(2),'y'); y=A.gen(); print A
            Univariate Polynomial Ring in y over Finite field of size 2
            sage: B = A.quotient(y**2 + y + 1, 'y2'); print B
            Univariate Quotient Polynomial Ring in y2 over Finite field of size 2 with modulus y^2 + y + 1
            sage: C = PolynomialRing(B, 'x'); x=C.gen(); print C
            Univariate Polynomial Ring in x over Univariate Quotient Polynomial Ring in y2 over Finite field of size 2 with modulus y^2 + y + 1
            sage: R = C.quotient(x**3 - 5); print R
            Univariate Quotient Polynomial Ring in x over Univariate Quotient Polynomial Ring in y2 over Finite field of size 2 with modulus y^2 + y + 1 with modulus x^3 + 1
            

        Next we create a number field, but viewed as a quotient of a
        polynomial ring over $\Q$:
            sage: R = PolynomialRing(RationalField(), 'x'); x = R.gen()
            sage: S = R.quotient(x**3 + 2*x - 5, 'a')
            sage: S
            Univariate Quotient Polynomial Ring in a over Rational Field with modulus x^3 + 2*x - 5
            sage: S.is_field()
            True
            sage: S.degree()
            3

        There are conversion functions for easily going back and forth
        between quotients of polynomial rings over $\Q$ and number
        fields:
            sage: K = S.number_field(); K
            Number Field in a with defining polynomial x^3 + 2*x - 5
            sage: K.polynomial_quotient_ring()
            Univariate Quotient Polynomial Ring in a over Rational Field with modulus x^3 + 2*x - 5
            
        """
        if not isinstance(R, polynomial_ring.PolynomialRing):
            raise TypeError, "R must be a univariate polynomial ring."

        if not isinstance(f, polynomial.Polynomial):
            raise TypeError, "f must be a Polynomial"
        
        if f.parent() != R:
            raise TypeError, "f must have parent R"

        if variable is None:
            variable = R.variable()

        self.__R = R
        self.__f = f
        self.__variable = variable
        ring.Ring.__init__(self)

    def __call__(self, x):
        """
        Coerce x into this quotient ring.  Anything that can be
        coerced into the polynomial ring can be coerced into the
        quotient.

        INPUT:
            x -- object to be coerced

        OUTPUT:
            PolynomialQuotientRingElement -- resulting of coercing x
            into this ring.

        EXAMPLES:
            sage: R = PolynomialRing(RationalField(), 'x'); x = R.gen()
            sage: S = R.quotient(x**3-3*x+1, 'alpha')
            sage: S(x)
            alpha
            sage: S(x**3)
            3*alpha - 1
            sage: S([1,2])
            2*alpha + 1
            sage: S([1,2,3,4,5])
            18*alpha^2 + 9*alpha - 3
            sage: S(S.gen()+1)
            alpha + 1
            sage: S(S.gen()**10+1)
            90*alpha^2 - 109*alpha + 28
        """
        if isinstance(x, PolynomialQuotientRingElement):
            if x.parent() == self:
                return x
        return PolynomialQuotientRingElement(self, self.__R(x) % self.__f, check=False)

    def __cmp__(self, other):
        """
        Compare self and other.

        EXAMPLES:
            sage: Rx = PolynomialRing(RationalField(), 'x'); x = Rx.gen()
            sage: Ry = PolynomialRing(RationalField(), 'y'); y = Ry.gen()
            sage: Rx == Ry
            False
            sage: Qx = Rx.quotient(x**2+1)
            sage: Qy = Ry.quotient(y**2+1)
            sage: Qx == Qy
            False
            sage: Qx == Qx
            True
            sage: Qz = Rx.quotient(x**2+1)
            sage: Qz == Qx
            True
        """
        if not isinstance(other, PolynomialQuotientRing):
            return -1
        if self.polynomial_ring() != other.polynomial_ring():
            return -1
        return self.modulus().__cmp__(other.modulus())

    def __contains__(self, x):
        """
        Returns True if this ring contains x, i.e., x is a polynomial
        quotient element with parent this ring.

        EXAMPLES:
            sage: R = PolynomialRing(RationalField(), 'x'); x = R.gen()
            sage: S = R.quotient(x**3-3*x+1, 'alpha')
            sage: S.gen()**2 in S
            True
            sage: x in S
            False

        Note that no coercion is done before checking, so the integer 1
        is not in S, but S(1) is.
            sage: 1 in S
            False
            sage: a = S(1);  a
            1
            sage: a in S
            True
        """
        return isinstance(x, PolynomialQuotientRingElement) and \
               x.parent() == self

    def __repr__(self):
        return "Univariate Quotient Polynomial Ring in %s over %s with modulus %s"%(
            self.variable(), self.base_ring(), self.modulus())


    def base_ring(self):
        r"""
        Return the base base ring of the polynomial ring, of which
        this ring is a quotient.

        EXAMPLES:
        The base ring of $\Z[z]/(z^3 + z^2 + z + 1)$ is $\Z$.
            sage: R = PolynomialRing(IntegerRing(), 'z'); z = R.gen()
            sage: S = R.quotient(z**3 + z**2 + z + 1, 'beta')
            sage: S.base_ring()
            Integer Ring

        Next we make a polynomial quotient ring over $S$ and ask for its basering.
            sage: T = PolynomialRing(S); W = T.quotient(T.gen()**99 + 99)
            sage: W.base_ring()
            Univariate Quotient Polynomial Ring in beta over Integer Ring with modulus z^3 + z^2 + z + 1
        """
        return self.__R.base_ring()

    def characteristic(self):
        """
        Return the characteristic of this quotient ring.

        This is always the same as the characteristic of the base ring.

        EXAMPLES:
            sage: R = PolynomialRing(IntegerRing(), 'z'); z = R.gen()
            sage: S = R.quotient(z - 19, 'a')
            sage: S.characteristic()
            0
            sage: R = PolynomialRing(GF(9), 'x'); x = R.gen()
            sage: S = R.quotient(x**3 + 1)
            sage: S.characteristic()
            3
        """
        return self.base_ring().characteristic()

    def degree(self):
        """
        Return the degree of this quotient ring.  The degree is the
        degree of the polynomial that we quotiented out by.

        EXAMPLES:
            sage: R = PolynomialRing(GF(3), 'x'); x = R.gen()
            sage: S = R.quotient(x**2005 + 1)
            sage: S.degree()
            2005
        """
        return self.modulus().degree()

    def discriminant(self, v=None):
        """
        Return the discriminant of this ring over the base ring.  This
        is by definition the discriminant of the polynomial that we
        quotiented out by.

        EXAMPLES:
            sage: R = PolynomialRing(RationalField(), 'x'); x = R.gen()
            sage: S = R.quotient(x**3 + x**2 + x + 1)
            sage: S.discriminant()
            -16
            sage: S = R.quotient((x + 1) * (x + 1))
            sage: S.discriminant()
            0

        The discriminant of the quotient polynomial ring need not
        equal the discriminant of the corresponding number field,
        since the discriminant of a number field is by definition the
        discriminant of the ring ring of integers of the number field:
            sage: S = R.quotient(x**2 - 8)
            sage: S.number_field().discriminant()
            8
            sage: S.discriminant()
            32
        """
        return self.modulus().discriminant()

    def gen(self, n=0):
        """
        Return the generator of this quotient ring.  This is the
        equivalence class of the image of the generator of the
        polynomial ring.

        EXAMPLES:
            sage: R = PolynomialRing(RationalField(), 'x'); x = R.gen()
            sage: S = R.quotient(x**2 - 8, 'gamma')
            sage: S.gen()
            gamma
        """
        if n != 0:
            raise IndexError, "Only one generator."
        try:
            return self.__gen
        except:
            self.__gen = self(self.polynomial_ring().gen())
            return self.__gen

    def is_field(self):
        """
        Return whether or not this quotient ring is a field.

        EXAMPLES:
            sage: R = PolynomialRing(IntegerRing(), 'z'); z = R.gen()
            sage: S = R.quotient(z**2-2)
            sage: S.is_field()
            False
            sage: R = PolynomialRing(RationalField(), 'x'); x = R.gen()
            sage: S = R.quotient(x**2 - 2)
            sage: S.is_field()
            True
        """
        return self.base_ring().is_field() and self.modulus().is_irreducible()

    def modulus(self):
        """
        Return the polynomial modulus of this quotient ring.

        EXAMPLES:
            sage: R = PolynomialRing(GF(3), 'x'); x = R.gen()
            sage: S = R.quotient(x**2 - 2)
            sage: S.modulus()
            x^2 + 1
        """
        return self.__f

    def name(self):
        """
        Return a short name useful in caching this ring. (Mainly for internal use.)

        EXAMPLES:
            sage: R = PolynomialRing(GF(3), 'x'); x = R.gen()
            sage: S = R.quotient(x**2 - 2)
            sage: S.name()
            'polyquox^2+1'
        """
        x="polyquo%s"%self.modulus()
        x=x.replace(" ","").replace("*","")
        return x

    def ngens(self):
        """
        Return the number of generators of this quotient ring over the
        base ring.  This function always returns 1.

        EXAMPLES:
            sage: R = PolynomialRing(RationalField(), 'x'); x = R.gen()
            sage: S = PolynomialRing(R, 'y'); y=S.gen()
            sage: T = S.quotient(y + x, 'z')
            sage: T
            Univariate Quotient Polynomial Ring in z over Univariate Polynomial Ring in x over Rational Field with modulus y + x
            sage: T.ngens()
            1
        """
        return 1
        
    def number_field(self):
        """
        Return the number field isomorphic to this quotient polynomial
        ring, if possible.

        EXAMPLES:
            sage: R = PolynomialRing(RationalField(), 'x'); x = R.gen()
            sage: S = R.quotient(x**29-17*x-1, 'alpha')
            sage: K = S.number_field()
            sage: K
            Number Field in alpha with defining polynomial x^29 - 17*x - 1
            sage: alpha = K.gen()
            sage: alpha**29
            17*alpha + 1
        """
        if self.characteristic() != 0:
            raise ArithmeticError, "Polynomial quotient ring is not isomorphic to a number field (it has positive characteristic)."
        
        if not isinstance(self.base_ring(), rational_field.RationalField):
            raise NotImplementedError, "Computation of number field only implemented for quotients of the polynomial ring over the rational field."
        return number_field.NumberField(self.modulus(), self.variable())

    def polynomial_ring(self):
        """
        Return the polynomial ring of which this ring is the quotient.

        EXAMPLES:
            sage: R = PolynomialRing(RationalField(), 'x'); x = R.gen()
            sage: S = R.quotient(x**2-2)
            sage: S.polynomial_ring()
            Univariate Polynomial Ring in x over Rational Field
        """
        return self.__R

    def variable(self, new_var=None):
        """
        Returns or sets the print representation of the generator of
        this quotient ring.

        If the optional string argument new_var is given, then the
        name of the variable is set to new_var.  If it is not given,
        this function returns the current print representation of the
        generator.
        
        INPUT:
            new_value -- (optional) string
            
        OUTPUT:
            str -- the name of the generator

        EXAMPLES:
            sage: R = PolynomialRing(RationalField(), 'x'); x = R.gen()
            sage: S = R.quotient(x**2-2, 'v')
            sage: S
            Univariate Quotient Polynomial Ring in v over Rational Field with modulus x^2 - 2
            sage: S.variable()
            'v'
            sage: S.variable('w')
            sage: S
            Univariate Quotient Polynomial Ring in w over Rational Field with modulus x^2 - 2
            sage: S.variable()
            'w'
        """
        if new_var is None:
            return self.__variable
        self.__variable = new_var


####################################################################
# Elements
####################################################################

class PolynomialQuotientRingElement(element.Element):
    """
    Element of a quotient of a polynomial ring.
    """
    def __init__(self, parent, x, check=True):
        """
        Create an element of the quotient of a polynomial ring.
        
        INPUT:
            parent -- a quotient of a polynomial ring
            x -- a polynomial
            check -- bool (optional): whether or not to verify 
                     that x is a valid element of the polynomial
                     ring and reduced (mod the modulus).
        """
        if check:
            if not isinstance(parent, PolynomialQuotientRing):
                raise TypeError, "parent must be a PolynomialQuotientRing"
            if not isinstance(x, polynomial.Polynomial):
                raise TypeError, "x must be a Polynomial"
            if not x in parent.polynomial_ring():
                raise TypeError, "x must be in the polynomial ring of the parent"
            if x.degree() > parent.degree():
                x %= parent.modulus()
        self.__x = x
        element.Element.__init__(self, parent)

    def __repr__(self):
        return self.__x.__repr__(self.parent().variable())

    def __add__(self, right):
        """
        Return the sum of two polynomial ring quotient elements.
        
        EXAMPLES:
            sage: R = PolynomialRing(RationalField(), 'x'); x = R.gen()
            sage: S = R.quotient(x**3-2, 'a'); a = S.gen()
            sage: (a**2 - 4) + (a+2)
            a^2 + a - 2
            sage: int(1) + a
            a + 1
        """
        if not isinstance(right, PolynomialQuotientRingElement) \
               or self.parent() != right.parent():
            return coerce.bin_op(self, right, operator.add)
        return PolynomialQuotientRingElement(self.parent(),
                               self.__x + right.__x, check=False)

    def __cmp__(self, other):
        """
        Compare this element with something else, where equality
        testing coerces the object on the right, if possible (and
        necessary).

        EXAMPLES:
            sage: R = PolynomialRing(RationalField(), 'x'); x = R.gen()
            sage: S = R.quotient(x**3-2, 'a'); a = S.gen()
            sage: S(1) == 1
            True
            sage: a**3 == 2
            True

        For the purposes of comparison in SAGE the quotient element
        $a^3$ is equal to $x^3$.  This is because when the comparison
        is performed, the right element is coerced into the parent of
        the left element, and $x^3$ coerces to $a^3$.

            sage: a == x
            True
            sage: a**3 == x**3
            True
            sage: x**3
            x^3
            sage: S(x**3)
            2

        There is no coercion from quotient elements to polynomials, so
        the other comparison is False:

            sage: x == a
            False
            
        """
        if not isinstance(other, PolynomialQuotientRingElement):
            return coerce.cmp(self, other)
        if self.parent() != other.parent():
            return -1
        return misc.generic_cmp(self.__x, other.__x)

    def __div__(self, right):
        """
        Return the quotient of two polynomial ring quotient elements.
        
        EXAMPLES:
            sage: R = PolynomialRing(RationalField(), 'x'); x = R.gen()
            sage: S = R.quotient(x**3-2, 'a'); a = S.gen()
            sage: (a**2 - 4) / (a+2)
            a - 2
        """
        if not isinstance(right, PolynomialQuotientRingElement) \
               or self.parent() != right.parent():
            return coerce.bin_op(self, right, operator.div)
        return self * ~right

    def __int__(self):
        """
        Coerce this element to an int if possible.

        EXAMPLES:
            sage: R = PolynomialRing(RationalField(), 'x'); x = R.gen()
            sage: S = R.quotient(x**3-2, 'a'); a = S.gen()
            sage: int(S(10))
            10
            sage: int(a)
            Traceback (most recent call last):
            ...
            ValueError: cannot coerce nonconstant polynomial to int
        """
        return int(self.__x)

    def __invert__(self):
        if self.__x.is_zero():
            raise ZeroDivisionError, \
               "element %s of quotient polynomial ring not invertible"%self
        g, _, a = self.parent().modulus().xgcd(self.__x)
        if g.degree() != 0:
            raise ZeroDivisionError, \
               "element %s of quotient polynomial ring not invertible"%self
        c = g[0]
        return PolynomialQuotientRingElement(self.parent(), (~c)*a, check=False)

    def __long__(self):
        """
        Coerce this element to a long if possible.

        EXAMPLES:
            sage: R = PolynomialRing(RationalField(), 'x'); x = R.gen()
            sage: S = R.quotient(x**3-2, 'a'); a = S.gen()
            sage: long(S(10))
            10
            sage: long(a)
            Traceback (most recent call last):
            ...
            ValueError: cannot coerce nonconstant polynomial to long
        """
        return long(self.__x)
    
    def __mul__(self, right):
        """
        Return the product of two polynomial ring quotient elements.
        
        EXAMPLES:
            sage: R = PolynomialRing(RationalField(), 'x'); x = R.gen()
            sage: S = R.quotient(x**3-2, 'a'); a = S.gen()
            sage: (a**2 - 4) * (a+2)
            2*a^2 - 4*a - 6
        """
        if not isinstance(right, PolynomialQuotientRingElement) \
               or self.parent() != right.parent():
            return coerce.bin_op(self, right, operator.mul)
        R = self.parent()
        prod = self.__x * right.__x
        g = prod % R.modulus()
        return PolynomialQuotientRingElement(R,g, check=False)

    def __neg__(self):
        return PolynomialQuotientRingElement(self.parent(), -self.__x)

    def __pow__(self, n):
        """
        Return a power of a polynomial ring quotient element.

        EXAMPLES:
            sage: R = PolynomialRing(Integers(9), 'x'); x = R.gen()
            sage: S = R.quotient(x**4 + 2*x**3 + x + 2, 'a'); a = S.gen()
            sage: a**100
            7*a^3 + 8*a + 7
        """
        n = int(n)
        if n < 0:
            x = self.__invert__()
            n *= -1
            return arith.generic_power(x, n, one=self.parent()(1))
        return arith.generic_power(self, n, one=self.parent()(1))

    def __radd__(self, left):
        return self.parent()(left) + self
    
    def __rsub__(self, left):
        return self.parent()(left) - self
    
    def __rmul__(self, left):
        return self.parent()(left) * self

    def __rdiv__(self, left):
        return self.parent()(left) / self        

    def __sub__(self, right):
        """
        Return the difference of two polynomial ring quotient elements.
        
        EXAMPLES:
            sage: R = PolynomialRing(RationalField(), 'x'); x = R.gen()
            sage: S = R.quotient(x**3-2, 'a'); a = S.gen()
            sage: (a**2 - 4) - (a+2)
            a^2 - a - 6
            sage: int(1) - a
            -a + 1
        """
        if not isinstance(right, PolynomialQuotientRingElement) \
               or self.parent() != right.parent():
            return coerce.bin_op(self, right, operator.sub)
        return PolynomialQuotientRingElement(self.parent(),
                               self.__x - right.__x, check=False)

    def charpoly(self):
        """
        The characteristic polynomial of this element, which is by
        definition the characteristic polynomial of right multiplication
        by this element.

        EXAMPLES:
            sage: R = PolynomialRing(RationalField(), 'x'); x = R.gen()
            sage: S = R.quotient(x**3 -389*x**2 + 2*x - 5, 'a'); a = S.gen()
            sage: a.charpoly()
            x^3 - 389*x^2 + 2*x - 5
        """
        return self.matrix().charpoly()

    def fcp(self):
        """
        Return the factorization of the characteristic polynomial
        of this element.

        EXAMPLES:
            sage: R = PolynomialRing(RationalField(), 'x'); x = R.gen()
            sage: S = R.quotient(x**3 -389*x**2 + 2*x - 5, 'a'); a = S.gen()
            sage: a.fcp()
            [(x^3 - 389*x^2 + 2*x - 5, 1)]
            sage: S(1).fcp()
            [(x - 1, 3)]
        """
        return self.charpoly().factor()

    def lift(self):
        """
        Return lift of this polynomial quotient ring element to the unique
        equivalent polynomial of degree less than the modulus.
        
        EXAMPLES:
            sage: R = PolynomialRing(RationalField(), 'x'); x = R.gen()
            sage: S = R.quotient(x**3-2, 'a'); a = S.gen()
            sage: b = a**2 - 3
            sage: b
            a^2 - 3
            sage: b.lift()
            x^2 - 3
        """
        return self.__x

    def list(self):
        """
        Return list of the elements of self, of length the same as
        the degree of the quotient polynomial ring.

        EXAMPLES:
            sage: R = PolynomialRing(RationalField(), 'x'); x = R.gen()
            sage: S = R.quotient(x**3 + 2*x - 5, 'a'); a = S.gen()
            sage: a**10
            -134*a^2 - 35*a + 300
            sage: (a**10).list()
            [300, -35, -134]
        """
        v = self.__x.list()
        R = self.parent()
        n = R.degree()
        return v + [R.base_ring()(0)]*(n - len(v))

    def matrix(self):
        """
        The matrix of right multiplication by this element on the
        power basis for the quotient ring.

        EXAMPLES:
            sage: R = PolynomialRing(RationalField(), 'x'); x = R.gen()
            sage: S = R.quotient(x**3 + 2*x - 5, 'a'); a = S.gen()
            sage: a.matrix()
            [ 0  1  0]
            [ 0  0  1]
            [ 5 -2  0]
        """
        # Mutiply each power of field generator on the right by this
        # element, then return the matrix whose rows are the
        # coefficients of the result.
        try:
            return self.__matrix
        except AttributeError:
            R = self.parent()
            v = []
            x = R.gen()
            a = R(1)
            d = R.degree()
            for n in range(d):
                v += (a*self).list()
                a *= x
            S = R.base_ring()
            import sage.linalg.linalg
            M = sage.linalg.linalg.MatrixSpace(S, d)
            self.__matrix = M(v)
            return self.__matrix

    def minpoly(self):
        """
        The minimal polynomial of this element, which is by definition
        the minimal polynomial of right multiplication by this
        element.
        """
        return self.matrix().minpoly()

    def norm(self):
        """
        The norm of this element, which is the norm of the matrix
        of right multiplication by this element.
        
        EXAMPLES:
            sage: R = PolynomialRing(RationalField(), 'x'); x = R.gen()
            sage: S = R.quotient(x**3 -389*x**2 + 2*x - 5, 'a'); a = S.gen()
            sage: a.norm()
            5
        """
        return self.matrix().determinant()

    def trace(self):
        """
        The trace of this element, which is the trace of the matrix
        of right multiplication by this element.

        EXAMPLES:
            sage: R = PolynomialRing(RationalField(), 'x'); x = R.gen()
            sage: S = R.quotient(x**3 -389*x**2 + 2*x - 5, 'a'); a = S.gen()
            sage: a.trace()
            389
        """
        return self.matrix().trace()

        
if __name__ ==  '__main__':
    import doctest, sys
    from sage.rings.rings import PolynomialRing, IntegerRing, RationalField, GF, Integers
    doctest.testmod(sys.modules[__name__])
